#!/usr/bin/env python
""" BIDS Terms -> NIDM-Exp Mappings
@author: David Keator <dbkeator@uci.edu>
"""
from . import Constants

# BIDS dataset_description -> NIDM constants mappings
dataset_description = {
    "BIDSVersion": Constants.BIDS["BIDSVersion"],
    "Name": Constants.NIDM_PROJECT_NAME,
    "Procedure": Constants.NIDM_PROJECT_DESCRIPTION,
    "License": Constants.NIDM_PROJECT_LICENSE,
    "ReferencesAndLinks": Constants.NIDM_PROJECT_REFERENCES,
    "Authors": Constants.NIDM_AUTHOR,
    "DatasetDOI": Constants.NIDM_DOI,
    "Funding": Constants.NIDM_FUNDING,
    "HowToAcknowledge": Constants.NIDM_ACKNOWLEDGEMENTS,
}

# BIDS Participants file -> NIDM constants mappings
participants = {
    "participant_id": Constants.NIDM_SUBJECTID
    # "sex" : Constants.NIDM_GENDER,
    # "age" : Constants.NIDM_AGE,
    # "gender" : Constants.NIDM_GENDER,
    # "diagnosis" : Constants.NIDM_DIAGNOSIS,
    # "handedness" : Constants.NIDM_HANDEDNESS
}
# scan metadata -> NIDM constants mappings
scans = {
    "anat": Constants.NIDM_MRI_ANATOMIC_SCAN,
    "func": Constants.NIDM_MRI_FUNCTION_SCAN,
    "dwi": Constants.NIDM_MRI_DWI_SCAN,
    "bval": Constants.NIDM_MRI_DWI_BVAL,
    "bvec": Constants.NIDM_MRI_DWI_BVEC,
    "T1w": Constants.NIDM_MRI_T1,
    "T2w": Constants.NIDM_MRI_T2,
    "inplaneT2": Constants.NIDM_MRI_T2,
    "bold": Constants.NIDM_MRI_FLOW,
    "dti": Constants.NIDM_MRI_DIFFUSION_TENSOR,
    "asl": Constants.NIDM_MRI_ASL,
}
# JSON file keys
json_keys = {
    # Image terms
    "run": Constants.NIDM_ACQUISITION_ENTITY,
    "ImageType": Constants.DICOM["ImageType"],
    "ManufacturerModelName": Constants.DICOM["ManufacturerModelName"],
    "Manufacturer": Constants.DICOM["Manufacturer"],
    "ScanningSequence": Constants.DICOM["ScanningSequence"],
    "SequenceVariant": Constants.DICOM["SequenceVariant"],
    "ScanOptions": Constants.DICOM["ScanOptions"],
    "MRAcquisitionType": Constants.DICOM["MRAcquisitionType"],
    "SequenceName": Constants.DICOM["SequenceName"],
    "RepetitionTime": Constants.DICOM["RepetitionTime"],
    "RepetitionTimePreparation": Constants.BIDS["RepetitionTimePreparation"],
    "ArterialSpinLabelingType": Constants.BIDS["ArterialSpinLabelingType"],
    "PostLabelingDelay": Constants.BIDS["PostLabelingDelay"],
    "BackgroundSuppression": Constants.BIDS["BackgroundSuppression"],
    "BackgroundSuppressionPulseTime": Constants.BIDS["BackgroundSuppressionPulseTime"],
    "BackgroundSuppressionNumberPulses": Constants.BIDS[
        "BackgroundSuppressionNumberPulses"
    ],
    "LabelingLocationDescription": Constants.BIDS["LabelingLocationDescription"],
    "LookLocker": Constants.BIDS["LookLocker"],
    "LabelingEfficiency": Constants.BIDS["LabelingEfficiency"],
    "LabelingDuration": Constants.BIDS["LabelingDuration"],
    "LabelingPulseAverageGradient": Constants.BIDS["LabelingPulseAverageGradient"],
    "LabelingPulseMaximumGradient": Constants.BIDS["LabelingPulseMaximumGradient"],
    "LabelingPulseDuration": Constants.BIDS["LabelingPulseDuration"],
    "LabelingPulseFlipAngle": Constants.BIDS["LabelingPulseFlipAngle"],
    "LabelingPulseInterval": Constants.BIDS["LabelingPulseInterval"],
    "PCASLType": Constants.BIDS["PCASLType"],
    "M0Type": Constants.BIDS["M0Type"],
    "TotalAcquiredPairs": Constants.BIDS["TotalAcquiredPairs"],
    "VascularCrushing": Constants.BIDS["VascularCrushing"],
    "EchoTime": Constants.BIDS["EchoTime"],
    "InversionTime": Constants.DICOM["InversionTime"],
    "NumberOfAverages": Constants.DICOM["NumberOfAverages"],
    "ImagingFrequency": Constants.DICOM["ImagingFrequency"],
    "MagneticFieldStrength": Constants.DICOM["MagneticFieldStrength"],
    "NumberOfPhaseEncodingSteps": Constants.DICOM["NumberOfPhaseEncodingSteps"],
    "EchoTrainLength": Constants.DICOM["EchoTrainLength"],
    "PercentSampling": Constants.DICOM["PercentSampling"],
    "PercentPhaseFieldOfView": Constants.DICOM["PercentPhaseFieldOfView"],
    "PixelBandwidth": Constants.DICOM["PixelBandwidth"],
    "AccelerationFactorPE": Constants.DICOM["AccelerationFactorPE"],
    "AccelNumReferenceLines": Constants.DICOM["AccelNumReferenceLines"],
    "TotalScanTimeSec": Constants.DICOM["TotalScanTimeSec"],
    "ReceiveCoilName": Constants.DICOM["ReceiveCoilName"],
    "DeviceSerialNumber": Constants.DICOM["DeviceSerialNumber"],
    "SoftwareVersions": Constants.DICOM["SoftwareVersions"],
    "ProtocolName": Constants.DICOM["ProtocolName"],
    "TransmitCoilName": Constants.DICOM["TransmitCoilName"],
    "AcquisitionMatrix": Constants.DICOM["AcquisitionMatrix"],
    "AcquisitionVoxelSize": Constants.BIDS["AcquisitionVoxelSize"],
    "InPlanePhaseEncodingDirection": Constants.DICOM["InPlanePhaseEncodingDirection"],
    "FlipAngle": Constants.BIDS["FlipAngle"],
    "VariableFlipAngleFlag": Constants.DICOM["VariableFlipAngleFlag"],
    "PatientPosition": Constants.DICOM["PatientPosition"],
    "PhaseEncodingDirection": Constants.BIDS["PhaseEncodingDirection"],
    "SliceTiming": Constants.BIDS["SliceTiming"],
    "TotalReadoutTime": Constants.BIDS["TotalReadoutTime"],
    "EffectiveEchoSpacing": Constants.NIDM["EffectiveEchoSpacing"],
    "NumberDiscardedVolumesByScanner": Constants.NIDM[
        "NumberDiscardedVolumesByScanner"
    ],
    "NumberDiscardedVolumesByUser": Constants.NIDM["NumberDiscardedVolumesByUser"],
    "DelayTime": Constants.NIDM["DelayTime"],
    "PulseSequenceType": Constants.DICOM["PulseSequenceName"],
    # Task Stuff
    "TaskName": Constants.NIDM_MRI_FUNCTION_TASK
    # "CogAtlasID" :
    # "CogPOID" :
    # "TaskDescription" :
    # "Instructions" :
    # "TaskFullName" :
    # "TaskName" :
}



import prov.model as pm
from .Acquisition import Acquisition
from ..core import Constants


class PETAcquisition(Acquisition):
    """
    Default constructor, creates a session activity and links to project object

    :param session: a session object

    """

    # constructor
    def __init__(self, session, attributes=None, uuid=None, add_default_type=True):
        """
        Default constructor, creates an acquisition object and links to acquisition activity object

        :param session: a session object
        :param attributes: optional attributes to add to entity
        :param uuid: optional uuid...used mostly for reading in existing NIDM document
        :return: none

        """
        # execute default parent class constructor
        super().__init__(session, attributes, uuid)
        # acquisition.graph._add_record(self)

        if add_default_type:
            self.add_attributes({pm.PROV_TYPE: Constants.NIDM_ACQUISITION_ACTIVITY})

        # carry graph object around
        self.graph = session.graph

    def __str__(self):
        return "NIDM-Experiment PET Acquisition Class"



import prov.model as pm
from .AcquisitionObject import AcquisitionObject
from ..core import Constants


class AssessmentObject(AcquisitionObject):
    """Class for NIDM-Experimenent generic AssessmentAcquisitionObject-Level Objects.

    Default constructor uses empty graph with namespaces added from NIDM/Scripts/Constants.py.
    Additional alternate constructors for user-supplied graphs and default namespaces (i.e. from Constants.py)
    and user-supplied graph and namespaces

    @author: David Keator <dbkeator@uci.edu>
    @copyright: University of California, Irvine 2017

    """

    # constructor
    def __init__(
        self,
        acquisition,
        assessment_type=None,
        attributes=None,
        uuid=None,
        add_default_type=True,
    ):
        """
        Default constructor, creates an acquisition object and links to acquisition activity object

        :param acquisition: a Acquisition activity object
        :param assessment_type: optional qualified name of assessment type (e.g. pm.QualifiedName(pm.Namespace("nidm",Constants.NIDM),"PositiveAndNegativeSyndromeScale"))
        :param attributes: optional attributes to add to entity
        :param uuid: optional uuid...used mostly for reading in existing NIDM document
        :return: none

        """
        # execute default parent class constructor
        super().__init__(acquisition, attributes, uuid)

        if add_default_type:
            self.add_attributes({pm.PROV_TYPE: Constants.NIDM_ASSESSMENT_ENTITY})
            self.add_attributes({pm.PROV_TYPE: Constants.NIDM_ACQUISITION_ENTITY})

        if assessment_type is not None:
            self.add_attributes({pm.PROV_TYPE: assessment_type})
        # carry graph object around
        self.graph = acquisition.graph

    def __str__(self):
        return "NIDM-Experiment Generic Assessment Object Class"



import collections
import functools
from rdflib import URIRef
from nidm.core import Constants
import nidm.experiment.CDE
from nidm.experiment.Query import (
    ACQUISITION_MODALITY,
    IMAGE_CONTRAST_TYPE,
    IMAGE_USAGE_TYPE,
    TASK,
    OpenGraph,
    URITail,
    expandUUID,
    getDataTypeInfo,
    matchPrefix,
    trimWellKnownURIPrefix,
)
from nidm.experiment.Utils import validate_uuid

isa = URIRef("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")
isPartOf = Constants.DCT["isPartOf"]
ValueType = collections.namedtuple(
    "ValueType",
    [
        "value",
        "label",
        "datumType",
        "hasUnit",
        "isAbout",
        "measureOf",
        "hasLaterality",
        "dataElement",
        "description",
        "subject",
        "project",
        "sourceVariable",
    ],
)
ActivityData = collections.namedtuple("ActivityData", ["category", "uuid", "data"])
QUERY_CACHE_SIZE = 0
BIG_CACHE_SIZE = 0


def makeValueType(
    value=None,
    label=None,
    datumType=None,
    hasUnit=None,
    isAbout=None,
    measureOf=None,
    hasLaterality=None,
    dataElement=None,
    description=None,
    subject=None,
    project=None,
    source_variable=None,
):
    return ValueType(
        str(value),
        str(label),
        str(datumType),
        str(hasUnit),
        str(isAbout),
        str(measureOf),
        str(hasLaterality),
        str(dataElement),
        str(description),
        str(subject),
        str(project),
        str(source_variable),
    )


def makeValueTypeFromDataTypeInfo(value, data_type_info_tuple):
    if not data_type_info_tuple:
        data_type_info_tuple = {}

    for key in [
        "label",
        "datumType",
        "hasUnit",
        "isAbout",
        "measureOf",
        "hasLaterality",
        "dataElement",
        "description",
        "subject",
        "project",
        "source_variable",
    ]:
        if key not in data_type_info_tuple:
            data_type_info_tuple[key] = None

    return ValueType(
        str(value),
        str(data_type_info_tuple["label"]),
        str(data_type_info_tuple["datumType"]),
        str(data_type_info_tuple["hasUnit"]),
        str(data_type_info_tuple["isAbout"]),
        str(data_type_info_tuple["measureOf"]),
        str(data_type_info_tuple["hasLaterality"]),
        str(data_type_info_tuple["dataElement"]),
        str(data_type_info_tuple["description"]),
        str(data_type_info_tuple["subject"]),
        str(data_type_info_tuple["project"]),
        str(data_type_info_tuple["source_variable"]),
    )


def expandID(id, namespace):  # noqa: A002
    """
    If the ID isn't a full URI already, make it one in the given namespace

    :param id:
    :param namespace:
    :return: full URI
    """
    if id.find("http") < 0:
        return namespace[id]
    # it has a http, but isn't a URIRef so convert it
    if isinstance(id, str):
        return URIRef(id)

    return id


@functools.lru_cache(maxsize=BIG_CACHE_SIZE)
def simplifyURIWithPrefix(nidm_file_tuples, uri):
    """
    Takes a URI and finds if there is a simple prefix for it in the graph
    :param rdf_graph:
    :param uri:
    :return: simple prefix or the original uri string
    """

    @functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
    def getNamespaceLookup(nidm_file_tuples):
        names = {}
        for f in nidm_file_tuples:
            rdf_graph = OpenGraph(f)
            for prefix, uri in rdf_graph.namespace_manager.namespaces():
                if str(uri) not in names:
                    names[str(uri)] = prefix
        return names

    names = getNamespaceLookup(tuple(nidm_file_tuples))
    # strip off the bit of URI after the last /
    trimed_uri = str(uri).split("/")[0:-1]
    trimed_uri = "/".join(trimed_uri) + "/"
    if trimed_uri in names:
        return names[trimed_uri]
    else:
        return uri


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getProjects(nidm_file_tuples):
    projects = []

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        # find all the sessions
        for project, _, _ in rdf_graph.triples((None, isa, Constants.NIDM["Project"])):
            projects.append(project)

    return projects


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSessions(nidm_file_tuples, project_id):
    project_uri = expandID(project_id, Constants.NIIRI)
    sessions = []

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        # find all the sessions
        for session, _, _ in rdf_graph.triples((None, isa, Constants.NIDM["Session"])):
            # check if it is part of our project
            if (session, isPartOf, project_uri) in rdf_graph:
                sessions.append(session)

    return sessions


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getAcquisitions(nidm_file_tuples, session_id):
    session_uri = expandID(session_id, Constants.NIIRI)
    acquisitions = []

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        # find all the sessions
        for acq, _, _ in rdf_graph.triples((None, isPartOf, session_uri)):
            # check if it is a acquisition
            if (acq, isa, Constants.NIDM["Acquisition"]) in rdf_graph:
                acquisitions.append(acq)

    return acquisitions


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSubject(nidm_file_tuples, acquisition_id):
    acquisition_uri = expandID(acquisition_id, Constants.NIIRI)

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        # find all the sessions
        for _, _, blank in rdf_graph.triples(
            (acquisition_uri, Constants.PROV["qualifiedAssociation"], None)
        ):
            for _, _, sub in rdf_graph.triples((blank, Constants.PROV["agent"], None)):
                if (
                    blank,
                    Constants.PROV["hadRole"],
                    Constants.SIO["Subject"],
                ) in rdf_graph:
                    return sub
    return None


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSubjects(nidm_file_tuples, project_id):
    subjects = set([])
    project_uri = expandID(project_id, Constants.NIIRI)

    sessions = getSessions(nidm_file_tuples, project_uri)
    for s in sessions:
        acquisitions = getAcquisitions(nidm_file_tuples, s)
        for acq in acquisitions:
            sub = getSubject(nidm_file_tuples, acq)
            subjects.add(sub)
    return subjects


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSubjectUUIDsfromID(nidm_file_tuples, sub_id):
    uuids = []
    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)

        result = rdf_graph.triples((None, Constants.NDAR["src_subject_id"], None))
        for s, _, o in result:
            if str(o) == str(sub_id):
                uuids.append(URITail(s))

    return uuids


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getSubjectIDfromUUID(nidm_file_tuples, subject_uuid):
    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        id_generator = rdf_graph.objects(
            subject=subject_uuid, predicate=Constants.NDAR["src_subject_id"]
        )
        for id_ in id_generator:
            return id_
    return None


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def normalizeSingleSubjectToUUID(nidm_file_tuples, id):  # noqa: A002
    if len(getSubjectUUIDsfromID(nidm_file_tuples, id)) > 0:
        return getSubjectUUIDsfromID(nidm_file_tuples, id)[0]
    return id


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getActivities(nidm_file_tuples, subject_id):
    activities = set([])

    # if we were passed in a sub_id rather than a UUID, lookup the associated UUID. (we might get multiple!)
    if validate_uuid(URITail(subject_id)):
        sub_uris = [subject_id]
    else:
        sub_uris = getSubjectUUIDsfromID(nidm_file_tuples, subject_id)

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        for subject_uri in sub_uris:
            subject_uri = expandID(subject_uri, Constants.NIIRI)
            for blank_node in rdf_graph.subjects(
                predicate=Constants.PROV["agent"], object=subject_uri
            ):
                for activity in rdf_graph.subjects(
                    predicate=Constants.PROV["qualifiedAssociation"], object=blank_node
                ):
                    if (activity, isa, Constants.PROV["Activity"]) in rdf_graph:
                        activities.add(activity)
    return activities


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def isAStatCollection(nidm_file_tuples, uri):
    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        if (
            ((uri, isa, Constants.NIDM["FSStatsCollection"]) in rdf_graph)
            or ((uri, isa, Constants.NIDM["FSLStatsCollection"]) in rdf_graph)
            or ((uri, isa, Constants.NIDM["ANTSStatsCollection"]) in rdf_graph)
            or ((uri, isa, Constants.NIDM["DerivativeCollection"]) in rdf_graph)
        ):
            return True
    return False


# def getDataElementInfo(nidm_file_list, id):
#
#     uuid = expandID(id, Constants.NIIRI)
#
#     for file in nidm_file_list:
#         rdf_graph = OpenGraph(file)
#         if (uuid, isa, Constants.NIDM['DataElement']) in rdf_graph:
#             label = list(rdf_graph.objects(subject=uuid, predicate=Constants.RDFS['label'])) [0]
#             description = list(rdf_graph.objects(subject=uuid, predicate=Constants.DCT['description'])) [0]
#             isAbout = list(rdf_graph.objects(subject=uuid, predicate=Constants.NIDM['isAbout'])) [0]
#             return makeValueType(label=label, description=description, isAbout=isAbout)
#
#     return False


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getActivityData(nidm_file_tuples, acquisition_id):
    acquisition_uri = expandID(acquisition_id, Constants.NIIRI)
    result = []
    category = None

    for file in nidm_file_tuples:
        rdf_graph = OpenGraph(file)
        # find everything generated by the acquisition
        for data_object, _, _ in rdf_graph.triples(
            (None, Constants.PROV["wasGeneratedBy"], acquisition_uri)
        ):
            # make sure this is an acquisition object
            if (data_object, isa, Constants.NIDM["AcquisitionObject"]) in rdf_graph:
                category = "instrument"
                # iterate over all the items in the acquisition object
                for _, p, o in rdf_graph.triples((data_object, None, None)):
                    dti = getDataTypeInfo(rdf_graph, p)
                    if dti:
                        # there is a DataElement describing this predicate
                        value_type = makeValueTypeFromDataTypeInfo(
                            value=trimWellKnownURIPrefix(o), data_type_info_tuple=dti
                        )
                        result.append(value_type)
                    else:
                        # Don't know exactly what this is so just set a label and be done.
                        if (
                            data_object,
                            isa,
                            Constants.ONLI["assessment-instrument"],
                        ) in rdf_graph:
                            result.append(
                                makeValueType(
                                    value=trimWellKnownURIPrefix(o),
                                    label=simplifyURIWithPrefix(
                                        nidm_file_tuples, str(p)
                                    ),
                                )
                            )
                            # result[ simplifyURIWithPrefix(nidm_file_list, str(p)) ] = trimWellKnownURIPrefix(o)
                        else:
                            result.append(
                                makeValueType(
                                    value=trimWellKnownURIPrefix(o),
                                    label=URITail(str(p)),
                                )
                            )
                            # result[ URITail(str(p))] = trimWellKnownURIPrefix(o)

            # or maybe it's a stats collection
            elif isAStatCollection(nidm_file_tuples, data_object):
                category = "derivative"
                for _, p, o in rdf_graph.triples((data_object, None, None)):
                    cde = getDataTypeInfo(rdf_graph, p)
                    result.append(
                        makeValueTypeFromDataTypeInfo(
                            value=str(o), data_type_info_tuple=cde
                        )
                    )
                    # result[ URITail(str(p)) ] = str(o)

    return ActivityData(
        category=category, uuid=trimWellKnownURIPrefix(acquisition_uri), data=result
    )


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def GetProjectAttributes(nidm_files_tuple, project_id):
    result = {
        ACQUISITION_MODALITY: set([]),
        IMAGE_CONTRAST_TYPE: set([]),
        IMAGE_USAGE_TYPE: set([]),
        TASK: set([]),
    }

    project_uuid = expandUUID(project_id)

    for file in nidm_files_tuple:
        rdf_graph = OpenGraph(file)
        # find all the projects
        for project, _, _ in rdf_graph.triples((None, None, Constants.NIDM["Project"])):
            # check if it is our project
            if str(project) == str(project_uuid):
                # get all the basic data from the project
                for _, predicate, obj in rdf_graph.triples((project, None, None)):
                    result[matchPrefix(str(predicate))] = str(obj)

    # now drill into the acquisition objects to get some specific
    # elements: AcquisitionModality, ImageContrastType, ImageUsageType, Task
    sessions = getSessions(nidm_files_tuple, project_id)
    for s in sessions:
        acquisitions = getAcquisitions(nidm_files_tuple, s)
        for a in acquisitions:
            acq_obj = getActivityData(nidm_files_tuple, a)
            for de in acq_obj.data:
                if de.label == "hadAcquisitionModality":
                    result[ACQUISITION_MODALITY].add(de.value)
                if de.label == "hadImageContrastType":
                    result[IMAGE_CONTRAST_TYPE].add(de.value)
                if de.label == "hadImageUsageType":
                    result[IMAGE_USAGE_TYPE].add(de.value)
                if de.label == "Task":
                    result[TASK].add(de.value)

    # de-set-ify items so they will play nice with JSON later
    result[ACQUISITION_MODALITY] = list(result[ACQUISITION_MODALITY])
    result[IMAGE_CONTRAST_TYPE] = list(result[IMAGE_CONTRAST_TYPE])
    result[IMAGE_USAGE_TYPE] = list(result[IMAGE_USAGE_TYPE])
    result[TASK] = list(result[TASK])

    return result


@functools.lru_cache(maxsize=BIG_CACHE_SIZE)
def GetAllPredicates(nidm_files_tuple):
    pred_set = set()
    for file in nidm_files_tuple:
        rdf_graph = OpenGraph(file)
        predicates = rdf_graph.predicates()
        for p in predicates:
            pred_set.add(p)
    return pred_set


@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def GetDataelements(nidm_files_tuple):
    result = {"data_elements": {"uuid": [], "label": [], "data_type_info": []}}
    found_uris = set()

    for file in nidm_files_tuple:
        rdf_graph = OpenGraph(file)
        # find all the datatypes
        for de_uri in rdf_graph.subjects(
            predicate=isa, object=Constants.NIDM["DataElement"]
        ):
            if de_uri not in found_uris:  # don't add duplicates
                dti = getDataTypeInfo(rdf_graph, de_uri)
                result["data_elements"]["uuid"].append(str(dti["dataElementURI"]))
                result["data_elements"]["label"].append(str(dti["label"]))
                result["data_elements"]["data_type_info"].append(dti)
                found_uris.add(de_uri)
        # find all the datatypes
        for de_uri in rdf_graph.subjects(
            predicate=isa, object=Constants.NIDM["PersonalDataElement"]
        ):
            if de_uri not in found_uris:  # don't add duplicates
                dti = getDataTypeInfo(rdf_graph, de_uri)
                result["data_elements"]["uuid"].append(str(dti["dataElementURI"]))
                result["data_elements"]["label"].append(str(dti["label"]))
                result["data_elements"]["data_type_info"].append(dti)
                found_uris.add(de_uri)

    # now look for any of the CDEs
    all_predicates = GetAllPredicates(nidm_files_tuple)
    cde_graph = nidm.experiment.CDE.getCDEs()
    cde_types = cde_graph.subjects(
        predicate=Constants.RDFS["subClassOf"], object=Constants.NIDM["DataElement"]
    )
    cde_type_set = set()  # i.e. fs:DataElement
    known_cde_types = set()  # i.e. fs_003579
    for t in cde_types:
        cde_type_set.add(t)
        for s in cde_graph.subjects(predicate=isa, object=t):
            known_cde_types.add(s)

    for predicate in all_predicates:
        if predicate in known_cde_types:
            dti = getDataTypeInfo(cde_graph, predicate)
            result["data_elements"]["uuid"].append(str(dti["dataElementURI"]))
            result["data_elements"]["label"].append(str(dti["label"]))
            result["data_elements"]["data_type_info"].append(dti)

    return result


def GetDataelementDetails(nidm_files_tuple, dataelement):
    result = {}

    for file in nidm_files_tuple:
        rdf_graph = OpenGraph(file)
        for de_uri in rdf_graph.subjects(
            predicate=isa, object=Constants.NIDM["DataElement"]
        ):
            dti = getDataTypeInfo(rdf_graph, de_uri)

            # check if this is the correct one
            if dataelement not in [
                str(dti["label"]),
                str(dti["dataElement"]),
                str(dti["dataElementURI"]),
            ]:
                continue

            result.update(dti)
            result["inProjects"] = set()

            # figure out what project the dataelement was used in
            uri = dti["dataElementURI"]

            a_list = rdf_graph.subjects(predicate=uri)
            for a in a_list:  # a is an assessment / AcquisitionObject
                b_list = rdf_graph.objects(
                    subject=a, predicate=Constants.PROV["wasGeneratedBy"]
                )
                for b in b_list:  # b is an Acquisition / Activity
                    c_list = rdf_graph.objects(
                        subject=b, predicate=Constants.DCT["isPartOf"]
                    )
                    for c in c_list:  # c is a session
                        d_list = rdf_graph.objects(
                            subject=c, predicate=Constants.DCT["isPartOf"]
                        )
                        for d in d_list:  # d is most likely a project
                            if d in rdf_graph.subjects(
                                predicate=isa, object=Constants.NIDM["Project"]
                            ):
                                result["inProjects"].add(f"{d} ({file})")

            return result  # found it, we are done

    if not result:  # didn't find it yet, check the CDEs
        cde_graph = nidm.experiment.CDE.getCDEs()
        for de_uri in cde_graph.subjects(predicate=isa):
            dti = getDataTypeInfo(cde_graph, de_uri)

            # check if this is the correct one
            if dataelement not in [
                str(dti["label"]),
                str(dti["dataElement"]),
                str(dti["dataElementURI"]),
            ]:
                continue

            result.update(dti)
            result["inProjects"] = set()
            result["inProjects"].add("Common Data Element")

            for file in nidm_files_tuple:
                rdf_graph = OpenGraph(file)
                if result["dataElementURI"] in rdf_graph.predicates():
                    result["inProjects"].add(file)

            return result  # found it, we are done

    return result



"""Graphical visualisation support for provONE.model.

This module produces graphical visualisation for ProvONE graphs.
Requires pydot module and Graphviz.

References:

* pydot homepage: https://github.com/erocarrera/pydot
* Graphviz:       http://www.graphviz.org/
* DOT Language:   http://www.graphviz.org/doc/info/lang.html

.. moduleauthor:: Sanu Ann Abraham <sanuann@mit.edu>
"""

from datetime import datetime
from html import escape
from prov.model import (
    PROV_ACTIVITY,
    PROV_AGENT,
    PROV_ALTERNATE,
    PROV_ASSOCIATION,
    PROV_ATTRIBUTE_QNAMES,
    PROV_ATTRIBUTION,
    PROV_BUNDLE,
    PROV_COMMUNICATION,
    PROV_DELEGATION,
    PROV_DERIVATION,
    PROV_END,
    PROV_ENTITY,
    PROV_GENERATION,
    PROV_INFLUENCE,
    PROV_INVALIDATION,
    PROV_MEMBERSHIP,
    PROV_MENTION,
    PROV_SPECIALIZATION,
    PROV_START,
    PROV_USAGE,
    Identifier,
    ProvException,
    sorted_attributes,
)
import pydot
from .Constants import (
    PROVONE_ATTRIBUTE_QNAMES,
    PROVONE_CLTODESTP,
    PROVONE_DATA,
    PROVONE_DATALINK,
    PROVONE_DATAONLINK,
    PROVONE_DLTOINPORT,
    PROVONE_DLTOOUTPORT,
    PROVONE_HASINPORT,
    PROVONE_HASOUTPORT,
    PROVONE_HASSUBPROCESS,
    PROVONE_INPORTTODL,
    PROVONE_INPUTPORT,
    PROVONE_ISPARTOF,
    PROVONE_OUTPORTTODL,
    PROVONE_OUTPUTPORT,
    PROVONE_PROCESS,
    PROVONE_PROCESSEXEC,
    PROVONE_SEQCTRLLINK,
    PROVONE_SOURCEPTOCL,
    PROVONE_USER,
)

__author__ = "Sanu Ann Abraham"
__email__ = "sanuann@mit.edu"


# Visual styles for various elements (nodes) and relations (edges)
# see http://graphviz.org/content/attrs

# DOT_PROVONE_STYLE = DOT_PROV_STYLE
DOT_PROVONE_STYLE = {
    PROVONE_PROCESS: {
        "shape": "oval",
        "style": "filled",
        "fillcolor": "#FFFC87",
        "color": "#808080",
    },
    PROVONE_PROCESSEXEC: {
        "shape": "box",
        "style": "filled",
        "fillcolor": "#9FB1FC",
        "color": "#0000FF",
    },
    PROVONE_INPUTPORT: {
        "shape": "oval",
        "style": "filled",
        "fillcolor": "#FFFC87",
        "color": "#808080",
    },
    PROVONE_OUTPUTPORT: {
        "shape": "oval",
        "style": "filled",
        "fillcolor": "#FFFC87",
        "color": "#808080",
    },
    PROVONE_DATALINK: {
        "shape": "oval",
        "style": "filled",
        "fillcolor": "#FFFC87",
        "color": "#808080",
    },
    PROVONE_SEQCTRLLINK: {
        "shape": "oval",
        "style": "filled",
        "fillcolor": "#FFFC87",
        "color": "#808080",
    },
    PROVONE_USER: {"shape": "house", "style": "filled", "fillcolor": "#FED37F"},
    # PROVONE_WORKFLOW: PROV_ENTITY,
    PROVONE_DATA: {
        "shape": "oval",
        "style": "filled",
        "fillcolor": "#FFFC87",
        "color": "#808080",
    },
    PROVONE_HASINPORT: {"label": "hasInPort", "fontsize": "10.0"},
    PROVONE_HASOUTPORT: {"label": "hasOutPort", "fontsize": "10.0"},
    PROVONE_HASSUBPROCESS: {"label": "hasSubProcess", "fontsize": "10.0"},
    PROVONE_INPORTTODL: {"label": "inPortToDL", "fontsize": "10.0"},
    PROVONE_OUTPORTTODL: {"label": "outPortToDL", "fontsize": "10.0"},
    PROVONE_CLTODESTP: {"label": "CLtoDestP", "fontsize": "10.0"},
    PROVONE_SOURCEPTOCL: {"label": "sourcePToCL", "fontsize": "10.0"},
    PROVONE_DLTOOUTPORT: {"label": "DLToOutPort", "fontsize": "10.0"},
    PROVONE_DLTOINPORT: {"label": "DLToInPort", "fontsize": "10.0"},
    PROVONE_DATAONLINK: {"label": "dataOnLink", "fontsize": "10.0"},
    PROVONE_ISPARTOF: {"label": "isPartOf", "fontsize": "10.0"},
    # Generic node
    0: {
        "shape": "oval",
        "style": "filled",
        "fillcolor": "lightgray",
        "color": "dimgray",
    },
    # Elements
    PROV_ENTITY: {
        "shape": "oval",
        "style": "filled",
        "fillcolor": "#FFFC87",
        "color": "#808080",
    },
    PROV_ACTIVITY: {
        "shape": "box",
        "style": "filled",
        "fillcolor": "#9FB1FC",
        "color": "#0000FF",
    },
    PROV_AGENT: {"shape": "house", "style": "filled", "fillcolor": "#FED37F"},
    PROV_BUNDLE: {"shape": "folder", "style": "filled", "fillcolor": "aliceblue"},
    # Relations
    PROV_GENERATION: {
        "label": "wasGeneratedBy",
        "fontsize": "10.0",
        "color": "darkgreen",
        "fontcolor": "darkgreen",
    },
    PROV_USAGE: {
        "label": "used",
        "fontsize": "10.0",
        "color": "red4",
        "fontcolor": "red",
    },
    PROV_COMMUNICATION: {"label": "wasInformedBy", "fontsize": "10.0"},
    PROV_START: {"label": "wasStartedBy", "fontsize": "10.0"},
    PROV_END: {"label": "wasEndedBy", "fontsize": "10.0"},
    PROV_INVALIDATION: {"label": "wasInvalidatedBy", "fontsize": "10.0"},
    PROV_DERIVATION: {"label": "wasDerivedFrom", "fontsize": "10.0"},
    PROV_ATTRIBUTION: {
        "label": "wasAttributedTo",
        "fontsize": "10.0",
        "color": "#FED37F",
    },
    PROV_ASSOCIATION: {
        "label": "wasAssociatedWith",
        "fontsize": "10.0",
        "color": "#FED37F",
    },
    PROV_DELEGATION: {
        "label": "actedOnBehalfOf",
        "fontsize": "10.0",
        "color": "#FED37F",
    },
    PROV_INFLUENCE: {"label": "wasInfluencedBy", "fontsize": "10.0", "color": "grey"},
    PROV_ALTERNATE: {"label": "alternateOf", "fontsize": "10.0"},
    PROV_SPECIALIZATION: {"label": "specializationOf", "fontsize": "10.0"},
    PROV_MENTION: {"label": "mentionOf", "fontsize": "10.0"},
    PROV_MEMBERSHIP: {"label": "hadMember", "fontsize": "10.0"},
}
# DOT_PROVONE_STYLE = dict.fromkeys([PROVONE_PROCESS, PROVONE_PROCESSEXEC,
#                             PROVONE_INPUTPORT, PROVONE_OUTPUTPORT,
#                             PROVONE_DATALINK, PROVONE_SEQCTRLLINK, PROVONE_USER,
#                             PROVONE_DATA], {
#         'shape': 'oval', 'style': 'filled',
#         'fillcolor': '#FFFC87', 'color': '#808080'
#     })
# DOT_PROVONE_STYLE.update({
#     # Elements
#
# 	PROVONE_HASINPORT: {
# 		'label': 'hasInPort', 'fontsize': '10.0'
# 	}
#     })

ANNOTATION_STYLE = {
    "shape": "note",
    "color": "gray",
    "fontcolor": "black",
    "fontsize": "10",
}
ANNOTATION_LINK_STYLE = {"arrowhead": "none", "style": "dashed", "color": "gray"}
ANNOTATION_START_ROW = '<<TABLE cellpadding="0" border="0">'
ANNOTATION_ROW_TEMPLATE = """    <TR>
        <TD align=\"left\" href=\"%s\">%s</TD>
        <TD align=\"left\"%s>%s</TD>
    </TR>"""
ANNOTATION_END_ROW = "    </TABLE>>"


def htlm_link_if_uri(value):
    try:
        uri = value.uri
        return f'<a href="{uri}">{value}</a>'
    except AttributeError:
        return str(value)


def provone_to_dot(
    bundle,
    show_nary=True,
    use_labels=False,
    direction="BT",
    show_element_attributes=True,
    show_relation_attributes=True,
):
    """
    Convert a provenance bundle/document into a DOT graphical representation.

    :param bundle: The provenance bundle/document to be converted.
    :type bundle: :class:`ProvBundle`
    :param show_nary: shows all elements in n-ary relations.
    :type show_nary: bool
    :param use_labels: uses the prov:label property of an element as its name (instead of its identifier).
    :type use_labels: bool
    :param direction: specifies the direction of the graph. Valid values are "BT" (default), "TB", "LR", "RL".
    :param show_element_attributes: shows attributes of elements.
    :type show_element_attributes: bool
    :param show_relation_attributes: shows attributes of relations.
    :type show_relation_attributes: bool
    :returns:  :class:`pydot.Dot` -- the Dot object.
    """
    if direction not in {"BT", "TB", "LR", "RL"}:
        # Invalid direction is provided
        direction = "BT"  # reset it to the default value
    maindot = pydot.Dot(graph_type="digraph", rankdir=direction, charset="utf-8")

    node_map = {}
    count = [0, 0, 0, 0]  # counters for node ids

    def _bundle_to_dot(dot, bundle):
        def _attach_attribute_annotation(node, record):
            # Adding a node to show all attributes
            attributes = list(
                (attr_name, value)
                for attr_name, value in record.attributes
                if attr_name not in PROV_ATTRIBUTE_QNAMES
            )

            if not attributes:
                return  # No attribute to display

            # Sort the attributes.
            attributes = sorted_attributes(record.get_type(), attributes)

            ann_rows = [ANNOTATION_START_ROW]
            ann_rows.extend(
                ANNOTATION_ROW_TEMPLATE
                % (
                    attr.uri,
                    escape(str(attr)),
                    f' href="{value.uri}"' if isinstance(value, Identifier) else "",
                    escape(
                        str(value)
                        if not isinstance(value, datetime)
                        else str(value.isoformat())
                    ),
                )
                for attr, value in attributes
            )
            ann_rows.append(ANNOTATION_END_ROW)
            count[3] += 1
            annotations = pydot.Node(
                f"ann{count[3]}", label="\n".join(ann_rows), **ANNOTATION_STYLE
            )
            dot.add_node(annotations)
            dot.add_edge(pydot.Edge(annotations, node, **ANNOTATION_LINK_STYLE))

        def _add_bundle(bundle):
            count[2] += 1
            subdot = pydot.Cluster(
                graph_name=f"c{count[2]}", URL=f'"{bundle.identifier.uri}"'
            )
            if use_labels:
                if bundle.label == bundle.identifier:
                    bundle_label = f'"{bundle.label}"'
                else:
                    # Fancier label if both are different. The label will be
                    # the main node text, whereas the identifier will be a
                    # kind of subtitle.
                    bundle_label = (
                        f"<{bundle.label}<br />"
                        '<font color="#333333" point-size="10">'
                        f"{bundle.identifier}</font>>"
                    )
                subdot.set_label(f'"{bundle_label}"')
            else:
                subdot.set_label(f'"{bundle.identifier}"')
            _bundle_to_dot(subdot, bundle)
            dot.add_subgraph(subdot)
            return subdot

        def _add_node(record):
            count[0] += 1
            node_id = f"n{count[0]}"
            if use_labels:
                if record.label == record.identifier:
                    node_label = f'"{record.label}"'
                else:
                    # Fancier label if both are different. The label will be
                    # the main node text, whereas the identifier will be a
                    # kind of subtitle.
                    node_label = (
                        f"<{record.label}<br />"
                        '<font color="#333333" point-size="10">'
                        f"{record.identifier}</font>>"
                    )
            else:
                node_label = f'"{record.identifier}"'

            uri = record.identifier.uri
            print("record type: ", record.get_type())
            style = DOT_PROVONE_STYLE[record.get_type()]
            print("style: ", style)
            node = pydot.Node(node_id, label=node_label, URL=f'"{uri}"', **style)
            node_map[uri] = node
            dot.add_node(node)

            if show_element_attributes:
                _attach_attribute_annotation(node, rec)
            return node

        def _add_generic_node(qname):
            count[0] += 1
            node_id = f"n{count[0]}"
            node_label = f'"{qname}"'

            uri = qname.uri
            style = DOT_PROVONE_STYLE[0]
            node = pydot.Node(node_id, label=node_label, URL=f'"{uri}"', **style)
            node_map[uri] = node
            dot.add_node(node)
            return node

        def _get_bnode():
            count[1] += 1
            bnode_id = f"b{count[1]}"
            bnode = pydot.Node(bnode_id, label='""', shape="point", color="gray")
            dot.add_node(bnode)
            return bnode

        def _get_node(qname):
            if qname is None:
                return _get_bnode()
            uri = qname.uri
            if uri not in node_map:
                _add_generic_node(qname)
            return node_map[uri]

        records = bundle.get_records()
        relations = []
        for rec in records:
            if rec.is_element():
                _add_node(rec)
            else:
                # Saving the relations for later processing
                relations.append(rec)

        if not bundle.is_bundle():
            for b in bundle.bundles:
                _add_bundle(b)

        for rec in relations:
            args = rec.args
            # skipping empty records
            if not args:
                continue
            # picking element nodes
            nodes = [
                value
                for attr_name, value in rec.formal_attributes
                if attr_name in PROVONE_ATTRIBUTE_QNAMES
            ]
            other_attributes = [
                (attr_name, value)
                for attr_name, value in rec.attributes
                if attr_name not in PROV_ATTRIBUTE_QNAMES
            ]
            add_attribute_annotation = show_relation_attributes and other_attributes
            add_nary_elements = len(nodes) > 2 and show_nary
            style = DOT_PROVONE_STYLE[rec.get_type()]
            if len(nodes) < 2:  # too few elements for a relation?
                continue  # cannot draw this

            if add_nary_elements or add_attribute_annotation:
                # a blank node for n-ary relations or the attribute annotation
                bnode = _get_bnode()

                # the first segment
                dot.add_edge(
                    pydot.Edge(_get_node(nodes[0]), bnode, arrowhead="none", **style)
                )
                style = dict(style)  # copy the style
                del style["label"]  # not showing label in the second segment
                # the second segment
                dot.add_edge(pydot.Edge(bnode, _get_node(nodes[1]), **style))
                if add_nary_elements:
                    style["color"] = "gray"  # all remaining segment to be gray
                    for node in nodes[2:]:
                        if node is not None:
                            dot.add_edge(pydot.Edge(bnode, _get_node(node), **style))
                if add_attribute_annotation:
                    _attach_attribute_annotation(bnode, rec)
            else:
                # show a simple binary relations with no annotation
                dot.add_edge(
                    pydot.Edge(_get_node(nodes[0]), _get_node(nodes[1]), **style)
                )

    try:
        unified = bundle.unified()
    except ProvException:
        # Could not unify this bundle
        # try the original document anyway
        unified = bundle

    _bundle_to_dot(maindot, unified)
    return maindot



import prov.model as pm
from .Core import Core, getUUID
from ..core import Constants


class DataElement(pm.ProvEntity, Core):
    """Class for NIDM-Experiment DataElement Objects.

    Default constructor uses empty graph with namespaces added from NIDM/Scripts/Constants.py.
    Additional alternate constructors for user-supplied graphs and default namespaces (i.e. from Constants.py)
    and user-supplied graph and namespaces

    @author: David Keator <dbkeator@uci.edu>
    @copyright: University of California, Irvine 2019

    """

    # constructor
    def __init__(self, project, attributes=None, uuid=None, add_default_type=True):
        """
        Default constructor, creates an acquisition object and links to acquisition activity object

        :param project: NIDM project to add data element entity to.\
        :param attributes: optional attributes to add to entity
        :param uuid: optional uuid...used mostly for reading in existing NIDM document
        :return: none

        """
        self.graph = project.graph

        if uuid is None:
            # execute default parent class constructor
            super().__init__(
                project.graph,
                pm.QualifiedName(pm.Namespace("niiri", Constants.NIIRI), getUUID()),
                attributes,
            )
        else:
            self._uuid = uuid

            # check if niiri namespace is already defined
            niiri_ns = self.find_namespace_with_uri(uri=str(Constants.NIIRI))

            if niiri_ns is False:
                super().__init__(
                    project.graph,
                    pm.QualifiedName(
                        pm.Namespace("niiri", Constants.NIIRI), self.get_uuid()
                    ),
                    attributes,
                )
            else:
                super().__init__(
                    project.graph,
                    pm.QualifiedName(niiri_ns, self.get_uuid()),
                    attributes,
                )
        project.graph._add_record(self)

        if add_default_type:
            self.add_attributes({pm.PROV_TYPE: Constants.NIDM_DATAELEMENT})
        project.add_dataelements(self)

        # list to store acquisition objects associated with this activity
        self._derivative_objects = []
        # if constructor is called with a session object then add this acquisition to the session

    def __str__(self):
        return "NIDM-Experiment DataElement Class"



"""
This program will load in a CSV file and iterate over the header variable names
performing an elastic search of https://scicrunch.org/ for NIDM-ReproNim tagged
terms that fuzzy match the variable names.  The user will then interactively
pick a term to associate with the variable name.  The resulting annotated CSV
data will then be written to a NIDM data file.
"""

from argparse import ArgumentParser
from io import StringIO
import logging
import os
from os.path import basename, dirname, join
from shutil import copy2
import sys
import pandas as pd
from prov.model import Identifier, QualifiedName
from rdflib import RDF, Graph, Literal
from rdflib.namespace import split_uri
from nidm.core import Constants
from nidm.experiment import (
    AssessmentAcquisition,
    AssessmentObject,
    Derivative,
    DerivativeObject,
    Project,
    Session,
)
from nidm.experiment.Core import getUUID
from nidm.experiment.Query import (
    GetAcquisitionEntityFromSubjectSessionRun,
    GetAcquisitionEntityFromSubjectSessionTask,
    GetAcquisitionEntityFromSubjectSessionTaskRun,
    GetParticipantIDs,
    GetParticipantSessionsMetadata,
    GetParticipantUUIDFromSubjectID,
)
from nidm.experiment.Utils import (
    add_attributes_with_cde,
    addGitAnnexSources,
    csv_dd_to_json_dd,
    map_variables_to_terms,
    read_nidm,
    redcap_datadictionary_to_json,
)

# adding click command line params interface outside main function


def ask_idfield(df):
    """
    This function will ask the user which column in the supplied csv file (df) is the id field because we were unable
    to automatically detect it.
    :param: df = data frame of supplied csv file to csv2nidm (i.e. args.csv_file loaded as dataframe)
    : return id_field column name
    """

    # if we couldn't find a subject ID field in column_to_terms, ask user
    option = 1
    for column in df.columns:
        print(f"{option}: {column}")
        option = option + 1
    selection = input("Please select the subject ID field from the list above: ")
    # Make sure user selected one of the options.  If not present user with selection input again
    while (not selection.isdigit()) or (int(selection) > int(option)):
        # Wait for user input
        selection = input("Please select the subject ID field from the list above: \t")
    id_field = df.columns[int(selection) - 1]

    return id_field


def detect_idfield(dd):
    """This function will attempt to find the id field in the supplied data dictionary
    :param: dd = data dictionary returned from map_variables_to_terms function in Utils.py
    :return: id field if found else None
    """

    # look at column_to_terms dictionary for NIDM URL for subject id  (Constants.NIDM_SUBJECTID)
    id_field = None
    for key, value in dd.items():
        if "isAbout" in value:
            for concept in value["isAbout"]:
                for isabout_key, isabout_value in concept.items():
                    if isabout_key in ("url", "@id"):
                        if isabout_value == Constants.NIDM_SUBJECTID._uri:
                            # get variable name from NIDM JSON file format:
                            # DD(source=assessment_name, variable=column)
                            id_field = (
                                key.split("variable")[1]
                                .split("=")[1]
                                .split(")")[0]
                                .lstrip("'")
                                .rstrip("'")
                            )

                            break
    return id_field


def find_session_for_subjectid(session_num, subjectid, nidm_file):
    """
    This function will search the supplied nidm_file for subjectid in agents and then get all sessions linked to this
    agent and search for a session with session number matching session_num.
    :param: session_num = string of session number searching for
    :param: subjectid = string subject id of subject interested in
    :param: nidm_file = NIDM file to search
    :return: session uuid containing session number for this subject
    """

    # get list of session objects we'll need to match up a derivative with a session / scan acquisition
    session_metadata = GetParticipantSessionsMetadata([nidm_file], subjectid)

    # find session number matching session provided in args.csv_file
    # initially set derivative_session to None in case we don't find it
    derivative_session = None
    for _, row in session_metadata.iterrows():
        # csv_row is the current row being processed.  Check the session_number against the ses column
        if row["p"] == Constants.BIDS["session_number"]:
            if str(row["o"]) == session_num:
                # found the session number so now get the acquisition entities linked to this session
                # to match up the 'task' and 'run' information so derived data is linked correctly
                derivative_session = row["session_uuid"]

    return derivative_session


def match_acquistion_task_run_from_session(
    subject_id, session_uuid, task, run, nidm_file
):
    """
    This function will use the supplied session_uuid from the NIDM file and find the acquisition entity with
    metadata matching the supplied task.
    :param: session_uuid = NIDM file session uuid to search for acquisitions associated with
    :param: task = string task name to search acquisition entity metadata for
    :param: run = string run number to search acquisition entity metadata for
    :param: nidm_file = NIDM file to search
    :return: Returns UUID of acquisition activity and entity matching task and run
    """

    derivative_acq_entity = None
    acquisition_act = None

    # if session_uuid, task, and run are all specified then we can simply find the acquisition entity directly
    # and be done otherwise we have to iteratively search
    if (session_uuid is not None) and (task is not None) and (run is not None):
        acq_entity = GetAcquisitionEntityFromSubjectSessionTaskRun(
            nidm_file_list=[nidm_file],
            subject_id=subject_id,
            session_uuid=session_uuid,
            run=run,
            task=task,
        )

        for _, acq in acq_entity.iterrows():
            derivative_acq_entity = acq["acq_entity"]
            acquisition_act = acq["acq_activity"]
            break

    # user supplied session_uuid, a run, and not task
    elif (session_uuid is not None) and (task is None) and (run is not None):
        acq_entity = GetAcquisitionEntityFromSubjectSessionRun(
            nidm_file_list=[nidm_file],
            subject_id=subject_id,
            session_uuid=session_uuid,
            run=run,
        )
        for _, acq in acq_entity.iterrows():
            derivative_acq_entity = acq["acq_entity"]
            acquisition_act = acq["acq_activity"]
            break

    # user supplied session_uuid, a task, and not run
    elif (session_uuid is not None) and (task is not None) and (run is None):
        acq_entity = GetAcquisitionEntityFromSubjectSessionTask(
            nidm_file_list=[nidm_file],
            subject_id=subject_id,
            session_uuid=session_uuid,
            task=task,
        )
        for _, acq in acq_entity.iterrows():
            derivative_acq_entity = acq["acq_entity"]
            acquisition_act = acq["acq_activity"]
            break

    # if session_uuid is None because user didn't supply a session number in the CSV file
    # implying there was no session number in the BIDS dataset for this task/run, we have to search across all
    # sessions to find the one that has an acquisition object that wasGenerated by an acquisition that
    # contains the task and/or run matching what the user
    # supplied.
    elif session_uuid is None:
        # get session_uuids and metadata for this subject_id
        session_act = GetParticipantSessionsMetadata([nidm_file], subject_id)
        # session_act = GetSessionUUID([nidm_file])

        # for session, get linked acquisition entities and search for the supplied
        # task and run
        for _, session in session_act.iterrows():
            # if session_uuid, task, and run are all specified then we can simply find the acquisition entity directly
            # and be done otherwise we have to iteratively search
            if (task is not None) and (run is not None):
                acq_entity = GetAcquisitionEntityFromSubjectSessionTaskRun(
                    nidm_file_list=[nidm_file],
                    subject_id=subject_id,
                    session_uuid=session["session_uuid"],
                    run=run,
                    task=task,
                )

                for _, acq in acq_entity.iterrows():
                    derivative_acq_entity = acq["acq_entity"]
                    acquisition_act = acq["acq_activity"]
                    break

            # user supplied session_uuid, a run, and not task
            elif (task is None) and (run is not None):
                acq_entity = GetAcquisitionEntityFromSubjectSessionRun(
                    nidm_file_list=[nidm_file],
                    subject_id=subject_id,
                    session_uuid=session["session_uuid"],
                    run=run,
                )
                for _, acq in acq_entity.iterrows():
                    derivative_acq_entity = acq["acq_entity"]
                    acquisition_act = acq["acq_activity"]
                    break

            # user supplied session_uuid, a task, and not run
            elif (task is not None) and (run is None):
                acq_entity = GetAcquisitionEntityFromSubjectSessionTask(
                    nidm_file_list=[nidm_file],
                    subject_id=subject_id,
                    session_uuid=session["session_uuid"],
                    task=task,
                )
                for _, acq in acq_entity.iterrows():
                    derivative_acq_entity = acq["acq_entity"]
                    acquisition_act = acq["acq_activity"]
                    break

    return derivative_acq_entity, acquisition_act


def csv2nidm_main(args=None):
    if args is None:
        parser = ArgumentParser(
            description="This program will load in a CSV file and iterate over the header \
         variable names performing an elastic search of https://scicrunch.org/ for NIDM-ReproNim \
         tagged terms that fuzzy match the variable names.  The user will then interactively pick \
         a term to associate with the variable name.  The resulting annotated CSV data will \
         then be written to a NIDM data file.  Note, you must obtain an API key to Interlex by signing up \
         for an account at scicrunch.org then going to My Account and API Keys.  Then set the environment \
         variable INTERLEX_API_KEY with your key."
        )

        parser.add_argument(
            "-csv",
            dest="csv_file",
            required=True,
            help="Full path to CSV file to convert",
        )
        # parser.add_argument('-ilxkey', dest='key', required=True, help="Interlex/SciCrunch API key to use for query")
        dd_group = parser.add_mutually_exclusive_group()
        dd_group.add_argument(
            "-json_map",
            dest="json_map",
            required=False,
            help="Full path to user-supplied JSON file containing variable-term mappings.",
        )
        dd_group.add_argument(
            "-csv_map",
            dest="csv_map",
            required=False,
            help="Full path to user-supplied CSV-version of data dictionary containing the following "
            "required columns: "
            "source_variable, "
            "label, "
            "description, "
            "valueType, "
            "measureOf, "
            "isAbout(For multiple isAbout entries, use a ';' to separate them in a single column "
            "within the csv file dataframe), "
            "unitCode, "
            "minValue, "
            "maxValue, ",
        )
        dd_group.add_argument(
            "-redcap",
            dest="redcap",
            required=False,
            help="Full path to a user-supplied RedCap formatted data dictionary for csv file. ",
        )
        parser.add_argument(
            "-nidm",
            dest="nidm_file",
            required=False,
            help="Optional full path of NIDM file to add CSV->NIDM converted graph to ",
        )
        parser.add_argument(
            "-no_concepts",
            action="store_true",
            required=False,
            help="If this flag is set then no concept associations will be "
            "asked of the user.  This is useful if you already have a -json_map specified without concepts and want to "
            "simply run this program to get a NIDM file with user interaction to associate concepts. ",
        )
        parser.add_argument(
            "-log",
            "--log",
            dest="logfile",
            required=False,
            default=None,
            help="full path to directory to save log file. Log file name is csv2nidm_[arg.csv_file].log ",
        )
        parser.add_argument(
            "-dataset_id",
            "--dataset_id",
            dest="dataset_identifier",
            required=False,
            default=None,
            help="If this is provided, which can be any dataset ID although its suggested to use a dataset "
            "DOI if available, unique data element IDs will use this information as part of the hash. ",
        )
        parser.add_argument(
            "-out",
            dest="output_file",
            required=False,
            help="Full path with filename to save NIDM file",
        )
        # added to support optional -derivative derived data group of arguments which includes the software metadata
        parser.add_argument(
            "-derivative",
            dest="derivative",
            required=False,
            help="If set, indicates CSV file provided is derivative data which includes columns 'ses','task','run' "
            "which will be used to identify the subject scan session, run, and verify against the task if an existing "
            "nidm file is provided and was made from bids (bidsmri2nidm). Otherwise these additional columns "
            "(ses, task,run) will be ignored. Further, adding a column source_url will provide links to the "
            "derived data.  After the -derivative parameter one must provide the software metadata "
            "CSV file which includes columns: title, description, version, url, cmdline, platform, ID "
            "These software metadata columns can have empty entries and are defined as follows: "
            "title: Title of the software"
            "description: Description of the software"
            "version: Version of the software"
            "url: Link to software"
            "cmdline: Command line used to run the software generating the results in the provided CSV "
            "ID: A url link to the term in a terminology resource (e.g. InterLex) for the software ",
        )
        args = parser.parse_args()

    # if we have a redcap datadictionary then convert it straight away to a json representation
    if args.redcap:
        json_map = redcap_datadictionary_to_json(args.redcap, basename(args.csv_file))
    elif args.json_map:
        json_map = args.json_map
    elif args.csv_map:
        if ".csv" in args.csv_map:
            # convert csv_map to a json version
            json_map = csv_dd_to_json_dd(args.csv_map)

        else:
            print("ERROR: -csv_map parameter must be a CSV file with .csv extension...")
            sys.exit(-1)
    else:
        json_map = None
    # open CSV file and load into
    # DBK added to accommodate TSV files with tab separator 3/15/21
    if args.csv_file.endswith(".csv"):
        df = pd.read_csv(args.csv_file)
    elif args.csv_file.endswith(".tsv"):
        df = pd.read_csv(args.csv_file, sep="\t", engine="python")
    else:
        print(
            "ERROR: input file must have .csv (comma-separated) or .tsv (tab separated) extensions/"
            "file types.  Please change your input file appropriately and re-run."
        )
        print("no NIDM file created!")
        sys.exit(-1)

    # added to require -out parameter if no -nidm parameter
    if args.nidm_file is None:
        if args.output_file is None:
            print(
                "ERROR: You must supply either an existing -nidm file to add "
                "metadata to or the -out output NIDM filename! "
            )
            parser.print_help()
            sys.exit(1)
        else:
            # set output file directory to location args.output_file
            output_dir = dirname(args.output_file)

    else:
        # set output file directory to location of existing NIDM file
        output_dir = dirname(args.nidm_file)

    # added to support CSV files that are derivatives representing processing of some scans
    if args.derivative:
        # first check if supplied CSV file in the -csv parameter has the derivative columns 'ses','task','run'
        if (
            ("ses" not in df.columns)
            or ("task" not in df.columns)
            or ("run" not in df.columns)
            or ("source_url" not in df.columns)
        ):
            print(
                "ERROR: -csv data file must have 'ses','task', 'run', 'source_url' columns (even if empty) when the "
                "-derivative parameter is provided.  See usage: "
            )
            parser.print_help()
            sys.exit(1)
        else:
            # remove -derivatives required columns 'ses','task','run' from CSV file
            # so map_variables_to_terms function doesn't complain about the columns not being annotated
            df = df.drop(["ses", "task", "run", "source_url"], axis=1)

            # load derivatives software metadata file and check it has the required columns
            if args.csv_file.endswith(".csv"):
                software_metadata = pd.read_csv(args.derivative)
            elif args.csv_file.endswith(".tsv"):
                software_metadata = pd.read_csv(
                    args.derivative, sep="\t", engine="python"
                )

            # check for required columns
            required_cols = [
                "title",
                "description",
                "version",
                "url",
                "cmdline",
                "platform",
                "ID",
            ]
            if not (set(required_cols).issubset(software_metadata.columns)):
                print(
                    "ERROR: -derivative software metadata file must contain columns title, description, "
                    "version, url, cmdline, platform, ID (even if empty).  See usage: "
                )
                parser.print_help()
                sys.exit(1)

    # code to associate variables in CSV file with data dictionary entries...otherwise go interactive in data dictionary
    # creation
    # if user did not specify -no_concepts then associate concepts interactively with user
    if not args.no_concepts:
        # if we're encoding a derivative then we'll want to collect the url of the software product to use for
        # cde prefixes
        if args.derivative:
            column_to_terms, cde = map_variables_to_terms(
                df=df,
                assessment_name=basename(args.csv_file),
                directory=output_dir,
                output_file=args.output_file,
                json_source=json_map,
                dataset_identifier=args.dataset_identifier,
                cde_namespace={
                    software_metadata["title"]
                    .to_string(index=False): software_metadata["url"]
                    .to_string(index=False)
                },
            )
        else:
            column_to_terms, cde = map_variables_to_terms(
                df=df,
                assessment_name=basename(args.csv_file),
                directory=output_dir,
                output_file=args.output_file,
                json_source=json_map,
                dataset_identifier=args.dataset_identifier,
            )
    # run without concept mappings
    else:
        # if we're encoding a derivative then we'll want to collect the url of the software product to use for
        # cde prefixes
        if args.derivative:
            column_to_terms, cde = map_variables_to_terms(
                df=df,
                assessment_name=basename(args.csv_file),
                directory=output_dir,
                output_file=args.output_file,
                json_source=json_map,
                associate_concepts=False,
                dataset_identifier=args.dataset_identifier,
                cde_namespace={
                    software_metadata["title"]
                    .to_string(index=False): software_metadata["url"]
                    .to_string(index=False)
                },
            )
        else:
            column_to_terms, cde = map_variables_to_terms(
                df=df,
                assessment_name=basename(args.csv_file),
                directory=output_dir,
                output_file=args.output_file,
                json_source=json_map,
                associate_concepts=False,
                dataset_identifier=args.dataset_identifier,
            )

    if args.logfile is not None:
        logging.basicConfig(
            filename=join(
                args.logfile,
                "csv2nidm_"
                + os.path.splitext(os.path.basename(args.csv_file))[0]
                + ".log",
            ),
            level=logging.DEBUG,
        )
        # add some logging info
        logging.info("csv2nidm %s", args)

    # If user has added an existing NIDM file as a command line parameter then add to existing file for subjects who exist in the NIDM file
    if args.nidm_file:
        if args.logfile:
            logging.info("Adding to NIDM file...")
        else:
            print("Adding to NIDM file...")
        # get subjectID list for later
        qres = GetParticipantIDs([args.nidm_file])

        # read in NIDM file
        if args.logfile:
            logging.info("Reading NIDM file...")
        else:
            print("Reading NIDM file...")
        project = read_nidm(args.nidm_file)
        # with open("/Users/dkeator/Downloads/test.ttl", "w", encoding="utf-8") as f:
        #    f.write(project.serializeTurtle())

        # get rdfs namespace for later use
        rdfs_ns = project.find_namespace_with_uri(
            uri="http://www.w3.org/2000/01/rdf-schema#"
        )
        # get nidm namespace for later use
        nidm_ns = project.find_namespace_with_uri(uri="http://purl.org/nidash/nidm#")

        id_field = detect_idfield(column_to_terms)

        # if we couldn't find a subject ID field in column_to_terms, ask user
        if id_field is None:
            # ask user for id field
            id_field = ask_idfield(df)

            # make sure id_field is a string for zero-padded subject ids
            # re-read data file with constraint that key field is read as string
            if args.csv_file.endswith(".csv"):
                df = pd.read_csv(args.csv_file, dtype={id_field: str})
            else:
                df = pd.read_csv(args.csv_file, dtype={id_field: str}, sep="\t")
        else:
            # make sure id_field is a string for zero-padded subject ids
            # re-read data file with constraint that key field is read as string
            df = pd.read_csv(args.csv_file, dtype={id_field: str})

        # ## use RDFLib here for temporary graph making query easier
        # rdf_graph = Graph()
        # rdf_graph.parse(source=StringIO(project.serializeTurtle()),format='turtle')

        # print("Querying for existing participants in NIDM graph....")

        # ## find subject ids and sessions in NIDM document
        # query = """SELECT DISTINCT ?session ?nidm_subj_id ?agent
        #            WHERE {
        #                ?activity prov:wasAssociatedWith ?agent ;
        #                    dct:isPartOf ?session  .
        #                ?agent rdf:type prov:Agent ;
        #                    ndar:src_subject_id ?nidm_subj_id .
        #            }"""
        # qres = rdf_graph.query(query)

        # variable to identify if any new data was added to the NIDM file, if not then we leave the existing nidm
        # file alone
        data_added = False

        # iterate over rows of csv file
        for _, df_row in df.iterrows():
            # see if this participant ID is in the NIDM file
            found_subject = False

            # search all subject ids in nidm file for one referenced in csv file
            for _, row in qres.iterrows():
                if (str(row.iloc[1]).lstrip("0") in df_row[id_field].lstrip("0")) or (
                    str(row.iloc[1]) in df_row[id_field]
                ):
                    found_subject = True
                    if args.logfile:
                        logging.info(
                            f"found participant {str(row.iloc[1]).lstrip('0')} in CSV file"
                        )
                    else:
                        print(
                            f"found participant {str(row.iloc[1]).lstrip('0')} in CSV file"
                        )
                    break

            # if the subject in CSV file isn't found in supplied nidm file then skip adding this derivative
            # information
            if not found_subject:
                continue

            data_added = True

            # find prov:Person associated with df_row[id_field]
            subject_uuid = GetParticipantUUIDFromSubjectID(
                nidm_file_list=[args.nidm_file],
                subject_id=df_row[id_field].lstrip("0"),
            )

            # added to support derivatives
            if args.derivative:
                # here we need to locate the session with bids:session_number equal to ses_task_run_df['ses'] for
                # this subject and then the acquisition with 'task' and optional 'run' so we can link the derivatives
                # to these data.  If all ('ses','run','task') are blank then we simply create a new session and
                # continue

                # get session number for this csv row
                session_num = df_row["ses"]

                # check if session_num is empty and if so, set to None
                # since we converted to a string we'll use string comparisons for 'nan'
                if str(session_num) == "nan":
                    session_num = None

                # now find session NIDM object for this subject
                derivative_session = find_session_for_subjectid(
                    session_num, str(df_row[id_field]).lstrip("0"), args.nidm_file
                )

                # get task from current csv row
                task = str(df_row["task"])

                # check if task is empty and if so, set to None
                if task == "nan":
                    task = None

                # get run from current csv row
                run = str(df_row["run"])

                # check if run is empty and if so, set to None
                if run == "nan":
                    run = None

                # now find acquisition entity matching the supplied task
                (
                    source_acq_entity,
                    source_activity,
                ) = match_acquistion_task_run_from_session(
                    subject_id=str(df_row[id_field]).lstrip("0"),
                    session_uuid=derivative_session,
                    task=task,
                    run=run,
                    nidm_file=args.nidm_file,
                )

                # check if we have a valid derivative_session, if so, use it.  If not, then skip this
                # derived entry

                if source_acq_entity is not None:
                    found_nm = project.find_namespace_with_uri(
                        software_metadata["url"].to_string(index=False)
                    )
                    if found_nm is False:
                        # add namespace for derived data software
                        project.addNamespace(
                            project.safe_string(
                                software_metadata["title"].to_string(index=False)
                            ),
                            software_metadata["url"].to_string(index=False),
                        )

                        found_nm = project.find_namespace_with_uri(
                            software_metadata["url"].to_string(index=False)
                        )

                    # create a derivative activity
                    der = Derivative(
                        project=project,
                    )

                    # create a derivative entity
                    der_entity = DerivativeObject(derivative=der)
                    der_entity.add_attributes(
                        {RDF.type: QualifiedName(nidm_ns, "DerivativeCollection")}
                    )

                    # add metadata to der_entity

                    # store other data from row with columns_to_term mappings
                    for row_variable in df_row.keys():
                        # check if row_variable is subject id, if so skip it
                        if (row_variable == id_field) or (
                            row_variable in ["ses", "task", "run", "subject_id"]
                        ):
                            continue
                        elif row_variable == "source_url":
                            der_entity.add_attributes(
                                {
                                    Constants.PROV["Location"]: Identifier(
                                        df_row["source_url"]
                                    )
                                }
                            )
                        else:
                            # check that the df_row[row_variable] contains some data/metadata, if so
                            # add to nidm file, if not skip it.
                            if str(df_row[row_variable]) != "nan":
                                add_attributes_with_cde(
                                    der_entity,
                                    cde,
                                    row_variable,
                                    Literal(df_row[row_variable]),
                                )
                    # link derivative activity to derivative_acq_entity with prov:used
                    namespace, name = split_uri(source_activity)

                    # find niiri namespace in project
                    niiri_ns = project.find_namespace_with_uri(str(Constants.NIIRI))

                    der.add_attributes(
                        {Constants.PROV["used"]: QualifiedName(niiri_ns, name)}
                    )

                    # add cmdline and platform to derivative activity
                    der.add_attributes(
                        {
                            software_metadata["url"].to_string(index=False)
                            + "cmdline": software_metadata["cmdline"].to_string(
                                index=False
                            ),
                            software_metadata["url"].to_string(index=False)
                            + "platform": software_metadata["platform"].to_string(
                                index=False
                            ),
                        }
                    )

                    # create software metadata agent

                    # find nidm namespace
                    nidm_ns = project.find_namespace_with_uri(str(Constants.NIDM))

                    software_agent = project.add_person(
                        attributes={
                            RDF["type"]: QualifiedName(nidm_ns, "SoftwareAgent")
                        },
                        add_default_type=False,
                    )

                    # add qualified association with subject

                    # find sio namespace in project
                    sio_ns = project.find_namespace_with_uri(str(Constants.SIO))

                    # if we need to add this namespace
                    if sio_ns is False:
                        # add sio namespace
                        project.addNamespace(prefix="sio", uri=str(Constants.SIO))

                        sio_ns = project.find_namespace_with_uri(str(Constants.SIO))

                    der.add_qualified_association(
                        person=subject_uuid["person_uuid"][0],
                        role=QualifiedName(sio_ns, "Subject"),
                    )

                    # add qualified association with software agent
                    # would prefer to use Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE here as the role
                    # but Constants.py has that as a rdflib Namespace but here we're adding data to a provDocument
                    # so using prov's QualifiedName and can't figure out how to convert rdflib Namespace to a prov
                    # qualified name...probably a matter of parsing the uri into two parts, one for prefix and the
                    # other for uri for prov QualifiedName function.
                    namespace, name = split_uri(
                        Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE
                    )
                    der.add_qualified_association(
                        person=software_agent,
                        role=QualifiedName(nidm_ns, name),
                    )
                    # add software metadata to software_agent
                    # uri:"http://ncitt.ncit.nih.gov/", prefix:"ncit", term:"age", value:15
                    # project.addAttributesWithNamespaces(software_agent,[{"uri":Constants.DCTYPES,
                    #                                "prefix": "dctypes", "term": "title","value":
                    #                                    software_metadata["title"].to_string(index=False)}])

                    # see if namespace for dcmitype exists, if not add it
                    # add namespaces to prov graph
                    dcmitype_ns = project.find_namespace_with_uri(
                        str(Constants.DCTYPES)
                    )

                    # if we need to add this namespace
                    if dcmitype_ns is False:
                        # add dcmitype namespace
                        project.addNamespace(
                            prefix="dcmitype", uri=str(Constants.DCTYPES)
                        )

                        dcmitype_ns = project.find_namespace_with_uri(
                            str(Constants.DCTYPES)
                        )

                    project.addAttributes(
                        software_agent,
                        {
                            QualifiedName(dcmitype_ns, "title"): software_metadata[
                                "title"
                            ].to_string(index=False)
                        },
                    )

                    # check if dct namespace needs to be added
                    dct_ns = project.find_namespace_with_uri(str(Constants.DCT))

                    # if we need to add this namespace
                    if dct_ns is False:
                        # add dcmitype namespace
                        project.addNamespace(prefix="dct", uri=str(Constants.DCT))

                        dct_ns = project.find_namespace_with_uri(str(Constants.DCT))

                    project.addAttributes(
                        software_agent,
                        {
                            QualifiedName(dct_ns, "description"): software_metadata[
                                "description"
                            ].to_string(index=False),
                            QualifiedName(dct_ns, "hasVersion"): software_metadata[
                                "version"
                            ].to_string(index=False),
                            QualifiedName(sio_ns, "URL"): software_metadata[
                                "url"
                            ].to_string(index=False),
                        },
                    )

                # if this isn't derivative data...
                if not args.derivative:
                    # add an assessment acquisition for the phenotype data to session and associate with agent
                    # acq=AssessmentAcquisition(session=nidm_session)

                    # create a new session for this assessment
                    new_session = Session(project=project)

                    acq = AssessmentAcquisition(session=new_session)
                    # add acquisition entity for assessment
                    acq_entity = AssessmentObject(acquisition=acq)
                    # add qualified association with existing agent
                    acq.add_qualified_association(
                        person=subject_uuid["person_uuid"][0],
                        role=Constants.NIDM_PARTICIPANT,
                    )

                    # add git-annex info if exists
                    num_sources = addGitAnnexSources(
                        obj=acq_entity,
                        filepath=args.csv_file,
                        bids_root=dirname(args.csv_file),
                    )
                    # if there aren't any git annex sources then just store the local directory information
                    if num_sources == 0:
                        # WIP: add absolute location of BIDS directory on disk for later finding of files
                        acq_entity.add_attributes(
                            {Constants.PROV["Location"]: "file:/" + args.csv_file}
                        )

                    # store file to acq_entity
                    acq_entity.add_attributes(
                        {Constants.NIDM_FILENAME: basename(args.csv_file)}
                    )

                    # store other data from row with columns_to_term mappings
                    for row_variable in df_row:
                        # check if row_variable is subject id, if so skip it
                        if row_variable == id_field:
                            continue
                        else:
                            if str(df_row[row_variable]) != "nan":
                                add_attributes_with_cde(
                                    acq_entity,
                                    cde,
                                    row_variable,
                                    Literal(df_row[row_variable]),
                                )

                    continue
        if args.logfile:
            logging.info("Adding CDEs to graph....")
        else:
            print("Adding CDEs to graph....")

        # with open(
        #    "/Users/dkeator/Downloads/before_cdes.ttl", "w", encoding="utf-8"
        # ) as f:
        #    f.write(project.serializeTurtle())
        # cde.serialize(destination="/Users/dkeator/Downloads/cdes.ttl", format="turtle")

        if data_added:
            # convert to rdflib Graph and add CDEs
            rdf_graph = Graph()
            rdf_graph.parse(source=StringIO(project.serializeTurtle()), format="turtle")
            rdf_graph = rdf_graph + cde

            if args.logfile:
                logging.info("Backing up original NIDM file...")
            else:
                print("Backing up original NIDM file...")
            copy2(src=args.nidm_file, dst=args.nidm_file + ".bak")
            if args.logfile:
                logging.info("Writing NIDM file....")
            else:
                print("Writing NIDM file....")
            rdf_graph.serialize(destination=args.nidm_file, format="turtle")
        else:
            if args.logfile:
                logging.info("No new data added, leaving existing nidm file alone...")
            else:
                print("No new data added, leaving existing nidm file alone...")

    else:
        if args.logfile:
            logging.info("Creating NIDM file...")
        else:
            print("Creating NIDM file...")
        # If user did not choose to add this data to an existing NIDM file then create a new one for the CSV data
        # create empty project
        project = Project()

        # add RDF namespace to project
        project.addNamespace(prefix="rdfs", uri="http://www.w3.org/2000/01/rdf-schema#")
        project.addNamespace(prefix="nidm", uri=str(Constants.NIDM))
        project.addNamespace(prefix="sio", uri=str(Constants.SIO))

        # look at column_to_terms dictionary for NIDM URL for subject id  (Constants.NIDM_SUBJECTID)
        id_field = detect_idfield(column_to_terms)

        if id_field is None:
            # ask user for id field
            id_field = ask_idfield(df)

            # make sure id_field is a string for zero-padded subject ids
            # re-read data file with constraint that key field is read as string
            if args.csv_file.endswith(".csv"):
                df = pd.read_csv(args.csv_file, dtype={id_field: str})
            else:
                df = pd.read_csv(args.csv_file, dtype={id_field: str}, sep="\t")
        else:
            # make sure id_field is a string for zero-padded subject ids
            # re-read data file with constraint that key field is read as string
            df = pd.read_csv(args.csv_file, dtype={id_field: str})

        # add namespace for derived data software
        project.addNamespace(
            project.safe_string(software_metadata["title"].to_string(index=False)),
            software_metadata["url"].to_string(index=False),
        )

        # get namespaces from document for use later....
        rdfs_ns = project.find_namespace_with_uri(
            "http://www.w3.org/2000/01/rdf-schema#"
        )
        nidm_ns = project.find_namespace_with_uri(str(Constants.NIDM))
        sio_ns = project.find_namespace_with_uri(str(Constants.SIO))

        # add a collection for storing project-level metadata
        provgraph = project.getGraph()
        collection = provgraph.collection(Constants.NIIRI[getUUID()])

        # simply add name of file to project metadata collection since we don't know anything about it
        collection.add_attributes({Constants.NIDM_FILENAME: args.csv_file})

        # iterate over rows and store in NIDM file
        for _, csv_row in df.iterrows():
            # added to support derivatives
            if args.derivative:
                # create a derivative activity
                der = Derivative(
                    project=project,
                )

                # create a derivative entity
                der_entity = DerivativeObject(derivative=der)
                der_entity.add_attributes(
                    {
                        QualifiedName(rdfs_ns, "type"): QualifiedName(
                            nidm_ns, "DerivativeCollection"
                        )
                    }
                )

                # add metadata to der_entity

                # store other data from row with columns_to_term mappings
                for row_variable, row_data in csv_row.items():
                    # check if row_variable is subject id, if so skip it
                    if (row_variable == id_field) or (
                        row_variable in ["ses", "task", "run"]
                    ):
                        continue
                    elif row_variable == "source_url":
                        der_entity.add_attributes(
                            {Constants.PROV["Location"]: Identifier(row_data)}
                        )
                    if str(row_data) != "nan":
                        # add data for this variable to derivative entity
                        add_attributes_with_cde(
                            der_entity,
                            cde,
                            row_variable,
                            Literal(row_data),
                        )

                # create subject agent
                subject_agent = project.add_person(
                    attributes=({Constants.NIDM_SUBJECTID: str(csv_row[id_field])})
                )

                # create software metadata agent
                software_agent = project.add_person(
                    attributes={
                        QualifiedName(rdfs_ns, "type"): QualifiedName(
                            nidm_ns, "SoftwareAgent"
                        )
                    },
                    add_default_type=False,
                )

                # add qualified association with subject
                der.add_qualified_association(
                    person=subject_agent,
                    role=QualifiedName(sio_ns, "Subject"),
                )

                found_nm = project.find_namespace_with_uri(
                    software_metadata["url"].to_string(index=False)
                )

                # add cmdline and platform to derivative activity
                der.add_attributes(
                    {
                        QualifiedName(found_nm, "cmdline"): software_metadata[
                            "cmdline"
                        ].to_string(index=False),
                        QualifiedName(found_nm, "platform"): software_metadata[
                            "platform"
                        ].to_string(index=False),
                    }
                )

                # add qualified association with software agent
                # would prefer to use Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE here as the role
                # but Constants.py has that as a rdflib Namespace but here we're adding data to a provDocument
                # so using prov's QualifiedName and can't figure out how to convert rdflib Namespace to a prov
                # qualified name...probably a matter of parsing the uri into two parts, one for prefix and the
                # other for uri for prov QualifiedName function.
                namespace, name = split_uri(
                    Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE
                )
                der.add_qualified_association(
                    person=software_agent,
                    role=QualifiedName(nidm_ns, name),
                )
                # add software metadata to software_agent
                # uri:"http://ncitt.ncit.nih.gov/", prefix:"ncit", term:"age", value:15
                # project.addAttributesWithNamespaces(software_agent,[{"uri":Constants.DCTYPES,
                #                                "prefix": "dctypes", "term": "title","value":
                #                                    software_metadata["title"].to_string(index=False)}])

                # add dctypes namespace

                # check if dct namespace needs to be added
                dcmitype_ns = project.find_namespace_with_uri(str(Constants.DCTYPES))

                # if we need to add this namespace
                if dcmitype_ns is False:
                    # add dcmitype namespace
                    project.addNamespace(prefix="dcmitype", uri=str(Constants.DCTYPES))

                    dcmitype_ns = project.find_namespace_with_uri(
                        str(Constants.DCTYPES)
                    )

                # project.addNamespace(prefix="dcmitype", uri=Constants.DCTYPES)
                project.addAttributes(
                    software_agent,
                    {
                        QualifiedName(dcmitype_ns, "title"): software_metadata[
                            "title"
                        ].to_string(index=False)
                    },
                )

                # check if dct namespace needs to be added
                dct_ns = project.find_namespace_with_uri(str(Constants.DCT))

                # if we need to add this namespace
                if dct_ns is False:
                    # add dcmitype namespace
                    project.addNamespace(prefix="dct", uri=str(Constants.DCT))

                    dct_ns = project.find_namespace_with_uri(str(Constants.DCT))

                project.addAttributes(
                    software_agent,
                    {
                        QualifiedName(dct_ns, "description"): software_metadata[
                            "description"
                        ].to_string(index=False),
                        QualifiedName(dct_ns, "hasVersion"): software_metadata[
                            "version"
                        ].to_string(index=False),
                        QualifiedName(sio_ns, "URL"): software_metadata[
                            "url"
                        ].to_string(index=False),
                    },
                )

            # not a derivative, assume an assessment
            else:
                # create a session object
                session = Session(project)

                # create and acquisition activity and entity
                acq = AssessmentAcquisition(session)
                acq_entity = AssessmentObject(acq)

                # add acq_entity to project collection
                provgraph.hadMember(collection, acq_entity)

                # create prov:Agent for subject
                # acq.add_person(attributes=({Constants.NIDM_SUBJECTID:row['participant_id']}))

                # add git-annex info if exists
                num_sources = addGitAnnexSources(
                    obj=acq_entity,
                    filepath=args.csv_file,
                    bids_root=os.path.dirname(args.csv_file),
                )
                # if there aren't any git annex sources then just store the local directory information
                if num_sources == 0:
                    # WIP: add absolute location of BIDS directory on disk for later finding of files
                    acq_entity.add_attributes(
                        {Constants.PROV["Location"]: "file:/" + args.csv_file}
                    )

                # store file to acq_entity
                acq_entity.add_attributes(
                    {Constants.NIDM_FILENAME: basename(args.csv_file)}
                )

                # store other data from row with columns_to_term mappings
                for row_variable, row_data in csv_row.iteritems():
                    if not row_data:
                        continue

                    # check if row_variable is subject id, if so skip it
                    if row_variable == id_field:
                        ### WIP: Check if agent already exists with the same ID.  If so, use it else create a new agent

                        # add qualified association with person
                        acq.add_qualified_association(
                            person=acq.add_person(
                                attributes=({Constants.NIDM_SUBJECTID: str(row_data)})
                            ),
                            role=Constants.NIDM_PARTICIPANT,
                        )

                        continue
                    else:
                        add_attributes_with_cde(
                            acq_entity, cde, row_variable, Literal(row_data)
                        )

                        # print(project.serializeTurtle())

        # convert to rdflib Graph and add CDEs
        # with open(
        #    "/Users/dkeator/Downloads/before_cdes.ttl", "w", encoding="utf-8"
        # ) as f:
        #    f.write(project.serializeTurtle())
        rdf_graph = Graph()
        rdf_graph.parse(source=StringIO(project.serializeTurtle()), format="turtle")
        rdf_graph = rdf_graph + cde

        if args.logfile:
            logging.info("Writing NIDM file....")
        else:
            print("Writing NIDM file....")
        # 5/7/25: added to accommodate the situation where user doesn't put .ttl at the end of the -out filename
        if ".ttl" not in args.output_file:
            output_file = args.output_file + ".ttl"
        else:
            output_file = args.output_file
        rdf_graph.serialize(destination=output_file, format="turtle")


if __name__ == "__main__":
    csv2nidm_main()



from __future__ import annotations
from dataclasses import InitVar, dataclass, field
from pathlib import Path
from types import TracebackType
from typing import IO, Any, Optional


@dataclass
class Reporter:
    output: Optional[IO[str]] = field(init=False)
    output_file: InitVar[str | Path | None]

    def __post_init__(self, output_file: str | Path | None) -> None:
        if output_file is not None:
            self.output = open(output_file, "w", encoding="utf-8")
        else:
            self.output = None

    def __enter__(self) -> Reporter:
        return self

    def __exit__(
        self,
        _exc_type: Optional[type[BaseException]],
        _exc_val: Optional[BaseException],
        _exc_tb: Optional[TracebackType],
    ) -> None:
        if self.output is not None:
            self.output.close()

    def print(self, *args: Any, end: str = "\n", sep: str = "") -> None:
        print(*args, end=end, sep=sep)
        if self.output is not None:
            print(*args, end=end, sep=sep, file=self.output)

    def print_file(self, *args: Any, end: str = "\n", sep: str = "") -> None:
        if self.output is not None:
            print(*args, end=end, sep=sep, file=self.output)



"""This program provides query functionality for NIDM-Experiment files"""

from json import dumps
import os
import sys
import click
from click_option_group import RequiredMutuallyExclusiveOptionGroup, optgroup
import pandas as pd
from nidm.experiment.CDE import getCDEs
from nidm.experiment.Query import (
    GetBrainVolumeDataElements,
    GetBrainVolumes,
    GetDataElements,
    GetInstrumentVariables,
    GetParticipantIDs,
    GetProjectInstruments,
    GetProjectsUUID,
    sparql_query_nidm,
)
from nidm.experiment.tools.click_base import cli
from nidm.experiment.tools.rest import RestParser


@cli.command()
@click.option(
    "--nidm_file_list",
    "-nl",
    required=True,
    help="A comma separated list of NIDM files with full path",
)
@click.option(
    "--cde_file_list",
    "-nc",
    required=False,
    help="A comma separated list of NIDM CDE files with full path. Can also be set in the CDE_DIR environment variable",
)
@optgroup.group(
    "Query Type",
    help="Pick among the following query type selections",
    cls=RequiredMutuallyExclusiveOptionGroup,
)
@optgroup.option(
    "--query_file",
    "-q",
    type=click.File("r"),
    help="Text file containing a SPARQL query to execute",
)
@optgroup.option(
    "--get_participants",
    "-p",
    is_flag=True,
    help="Parameter, if set, query will return participant IDs and prov:agent entity IDs",
)
@optgroup.option(
    "--get_instruments",
    "-i",
    is_flag=True,
    help="Parameter, if set, query will return list of onli:assessment-instrument:",
)
@optgroup.option(
    "--get_instrument_vars",
    "-iv",
    is_flag=True,
    help="Parameter, if set, query will return list of onli:assessment-instrument: variables",
)
@optgroup.option(
    "--get_dataelements",
    "-de",
    is_flag=True,
    help="Parameter, if set, will return all DataElements in NIDM file",
)
@optgroup.option(
    "--get_dataelements_brainvols",
    "-debv",
    is_flag=True,
    help="Parameter, if set, will return all brain volume DataElements in NIDM file along with details",
)
@optgroup.option(
    "--get_brainvols",
    "-bv",
    is_flag=True,
    help="Parameter, if set, will return all brain volume data elements and values along with participant IDs in NIDM file",
)
@optgroup.option(
    "--get_fields",
    "-gf",
    help="This parameter will return data for only the field names in the comma separated list (e.g. -gf age,fs_00003) from all nidm files supplied",
)
@optgroup.option("--uri", "-u", help="A REST API URI query")
@click.option(
    "--output_file",
    "-o",
    required=False,
    help="Optional output file (CSV) to store results of query",
)
@click.option(
    "-j/-no_j",
    required=False,
    default=False,
    help="Return result of a uri query as JSON",
)
@click.option(
    "--blaze",
    "-bg",
    required=False,
    help="Base URL for Blazegraph. Ex: http://172.19.0.2:9999/blazegraph/sparql",
)
@click.option(
    "-v",
    "--verbosity",
    required=False,
    help="Verbosity level 0-5, 0 is default",
    default="0",
)
def query(
    nidm_file_list,
    cde_file_list,
    query_file,
    output_file,
    get_participants,
    get_instruments,
    get_instrument_vars,
    get_dataelements,
    get_brainvols,
    get_dataelements_brainvols,
    get_fields,
    uri,
    blaze,
    j,
    verbosity,
):
    """
    This function provides query support for NIDM graphs.
    """
    # if there is a CDE file list, seed the CDE cache
    if cde_file_list:
        getCDEs(cde_file_list.split(","))

    if blaze:
        os.environ["BLAZEGRAPH_URL"] = blaze
        print(f"setting BLAZEGRAPH_URL to {blaze}")

    if get_participants:
        df = GetParticipantIDs(nidm_file_list.split(","), output_file=output_file)
        if (output_file) is None:
            print(df.to_string())

        return df
    elif get_instruments:
        # first get all project UUIDs then iterate and get instruments adding to output dataframe
        project_list = GetProjectsUUID(nidm_file_list.split(","))
        count = 1
        for project in project_list:
            if count == 1:
                df = GetProjectInstruments(
                    nidm_file_list.split(","), project_id=project
                )
                count += 1
            else:
                df = df.append(
                    GetProjectInstruments(nidm_file_list.split(","), project_id=project)
                )

        # write dataframe
        # if output file parameter specified
        if output_file is not None:
            df.to_csv(output_file)
            # with open(output_file,'w', encoding="utf-8") as myfile:
            #    wr=csv.writer(myfile,quoting=csv.QUOTE_ALL)
            #    wr.writerow(df)

            # pd.DataFrame.from_records(df,columns=["Instruments"]).to_csv(output_file)
        else:
            print(df.to_string())
    elif get_instrument_vars:
        # first get all project UUIDs then iterate and get instruments adding to output dataframe
        project_list = GetProjectsUUID(nidm_file_list.split(","))
        count = 1
        for project in project_list:
            if count == 1:
                df = GetInstrumentVariables(
                    nidm_file_list.split(","), project_id=project
                )
                count += 1
            else:
                df = df.append(
                    GetInstrumentVariables(
                        nidm_file_list.split(","), project_id=project
                    )
                )

        # write dataframe
        # if output file parameter specified
        if output_file is not None:
            df.to_csv(output_file)
        else:
            print(df.to_string())
    elif get_dataelements:
        datael = GetDataElements(nidm_file_list=nidm_file_list)
        # if output file parameter specified
        if output_file is not None:
            datael.to_csv(output_file)
        else:
            print(datael.to_string())
    elif get_fields:
        # fields only query.  We'll do it with the rest api
        restParser = RestParser(verbosity_level=int(verbosity))
        if output_file is not None:
            restParser.setOutputFormat(RestParser.OBJECT_FORMAT)
            df_list = []
        else:
            restParser.setOutputFormat(RestParser.CLI_FORMAT)
        # set up uri to do fields query for each nidm file
        for nidm_file in nidm_file_list.split(","):
            # get project UUID
            project = GetProjectsUUID([nidm_file])
            uri = (
                "/projects/"
                + project[0].toPython().split("/")[-1]
                + "?fields="
                + get_fields
            )
            # get fields output from each file and concatenate
            if output_file is None:
                # just print results
                print(restParser.run([nidm_file], uri))
            else:
                df_list.append(pd.DataFrame(restParser.run([nidm_file], uri)))

        if output_file is not None:
            # concatenate data frames
            df = pd.concat(df_list)
            # output to csv file
            df.to_csv(output_file)

    elif uri:
        restParser = RestParser(verbosity_level=int(verbosity))
        if j:
            restParser.setOutputFormat(RestParser.JSON_FORMAT)
        elif output_file is not None:
            restParser.setOutputFormat(RestParser.OBJECT_FORMAT)
        else:
            restParser.setOutputFormat(RestParser.CLI_FORMAT)
        df = restParser.run(nidm_file_list.split(","), uri)
        if output_file is not None:
            if j:
                with open(output_file, "w+", encoding="utf-8") as f:
                    f.write(dumps(df))
            else:
                # convert object df to dataframe and output
                pd.DataFrame(df).to_csv(output_file)
        else:
            print(df)

    elif get_dataelements_brainvols:
        brainvol = GetBrainVolumeDataElements(nidm_file_list=nidm_file_list)
        # if output file parameter specified
        if output_file is not None:
            brainvol.to_csv(output_file)
        else:
            print(brainvol.to_string())
    elif get_brainvols:
        brainvol = GetBrainVolumes(nidm_file_list=nidm_file_list)
        # if output file parameter specified
        if output_file is not None:
            brainvol.to_csv(output_file)
        else:
            print(brainvol.to_string())
    elif query_file:
        df = sparql_query_nidm(nidm_file_list.split(","), query_file, output_file)

        if (output_file) is None:
            print(df.to_string())

        return df
    else:
        print("ERROR: No query parameter provided.  See help:")
        print()
        os.system("pynidm query --help")
        sys.exit(1)


# it can be used calling the script `python nidm_query.py -nl ... -q ..
if __name__ == "__main__":
    query()



""" Tools for working with NIDM-Experiment files """

from os.path import basename, join, splitext
import click
from rdflib import Graph, util
from nidm.experiment.tools.click_base import cli


# adding click argument parsing
@cli.command()
@click.option(
    "--nidm_file_list",
    "-nl",
    required=True,
    help="A comma separated list of NIDM files with full path",
)
@click.option(
    "-t",
    "--type",
    "outtype",
    required=True,
    type=click.Choice(
        ["turtle", "jsonld", "xml-rdf", "n3", "trig"], case_sensitive=False
    ),
    help="If parameter set then NIDM file will be exported as JSONLD",
)
@click.option(
    "--outdir",
    "-out",
    required=False,
    help="Optional directory to save converted NIDM file",
)
def convert(nidm_file_list, outtype, outdir):
    """
    This function will convert NIDM files to various RDF-supported formats and name then / put them in the same
    place as the input file.
    """

    for nidm_file in nidm_file_list.split(","):
        # WIP: for now we use pynidm for jsonld exports to make more human readable and rdflib for everything
        # else.
        if outdir:
            outfile = join(outdir, splitext(basename(nidm_file))[0])
        else:
            outfile = join(splitext(nidm_file)[0])

        if outtype == "jsonld":
            ## read in nidm file
            # project = read_nidm(nidm_file)
            ## write jsonld file with same name
            # with open(outfile + ".json", "w", encoding="utf-8") as f:
            #    f.write(project.serializeJSONLD())
            graph = Graph()
            graph.parse(nidm_file, format=util.guess_format(nidm_file))
            graph.serialize(outfile + ".json", format="json-ld", indent=4)

        elif outtype == "turtle":
            ## graph = Graph()
            ## graph.parse(nidm_file, format=util.guess_format(nidm_file))
            ## graph.serialize(splitext(nidm_file)[0] + ".ttl", format='turtle')
            # project = read_nidm(nidm_file)
            # with open(outfile + ".ttl", "w", encoding="utf-8") as f:
            #    f.write(project.serializeTurtle())
            graph = Graph()
            graph.parse(nidm_file, format=util.guess_format(nidm_file))
            graph.serialize(outfile + ".ttl", format="turtle", indent=4)

        elif outtype == "xml-rdf":
            graph = Graph()
            graph.parse(nidm_file, format=util.guess_format(nidm_file))
            graph.serialize(outfile + ".xml", format="pretty-xml")
        elif outtype == "n3":
            graph = Graph()
            graph.parse(nidm_file, format=util.guess_format(nidm_file))
            graph.serialize(outfile + ".n3", format="n3")
        elif outtype == "trig":
            ## read in nidm file
            # project = read_nidm(nidm_file)
            # with open(outfile + ".trig", "w", encoding="utf-8") as f:
            #    f.write(project.serializeTrig())
            graph = Graph()
            graph.parse(nidm_file, format=util.guess_format(nidm_file))
            graph.serialize(outfile + ".trig", format="trig")

        else:
            print("Error, type is not supported at this time")


if __name__ == "__main__":
    convert()



from __future__ import annotations
import pytest
from nidm.experiment.tools.rest import RestParser
from nidm.util import urlretrieve

OPENNEURO_FILES = [
    "ds000002.nidm.ttl",
    "ds000003.nidm.ttl",
    "ds000011.nidm.ttl",
    "ds000017.nidm.ttl",
    "ds000101.nidm.ttl",
    "ds000108.nidm.ttl",
    "ds000113.nidm.ttl",
    "ds000114.nidm.ttl",
    "ds000120.nidm.ttl",
    "ds000122.nidm.ttl",
    "ds000138.nidm.ttl",
    "ds000171.nidm.ttl",
    "ds000208.nidm.ttl",
    "ds000214.nidm.ttl",
    "ds000222.nidm.ttl",
    "ds000224.nidm.ttl",
    "ds000238.nidm.ttl",
    "ds000246.nidm.ttl",
    "ds001021.nidm.ttl",
    "ds001178.nidm.ttl",
    "ds001232.nidm.ttl",
    "ds001241.nidm.ttl",
]

# OPENNEURO_FILES = ['ds000001.nidm.ttl',
#                    'ds000003.nidm.ttl']
#


@pytest.fixture(scope="module")
def openneuro_files(tmp_path_factory: pytest.TempPathFactory) -> list[str]:
    tmp_path = tmp_path_factory.mktemp("openneuro")
    for fname in OPENNEURO_FILES:
        dataset = fname.split(".")[0]
        urlretrieve(
            f"https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/openneuro/{dataset}/nidm.ttl",
            tmp_path / fname,
        )
    return [str(tmp_path / fname) for fname in OPENNEURO_FILES]


def test_dataelement_list(
    brain_vol_files: list[str], openneuro_files: list[str]
) -> None:
    rest_parser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    result = rest_parser.run(openneuro_files, "/dataelements")

    assert isinstance(result, dict)
    assert "data_elements" in result
    assert "uuid" in result["data_elements"]
    assert "label" in result["data_elements"]
    assert "data_type_info" in result["data_elements"]

    assert len(result["data_elements"]["label"]) != 0
    assert len(result["data_elements"]["label"]) == len(result["data_elements"]["uuid"])
    assert len(result["data_elements"]["label"]) == len(
        result["data_elements"]["data_type_info"]
    )

    for label in result["data_elements"]["label"]:
        assert label in [
            str(x["label"]) for x in result["data_elements"]["data_type_info"]
        ]

    for uuid in result["data_elements"]["uuid"]:
        assert uuid in [
            str(x["dataElementURI"]) for x in result["data_elements"]["data_type_info"]
        ]

    # now check for derivatives
    result = rest_parser.run(brain_vol_files, "/dataelements")
    assert isinstance(result, dict)
    assert (
        "Left-WM-hypointensities Volume_mm3 (mm^3)" in result["data_elements"]["label"]
    )


def test_dataelement_details(openneuro_files: list[str]) -> None:
    rest_parser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    # result = rest_parser.run(openneuro_files, '/dataelements')
    #
    # dti = rest_parser.run(openneuro_files, f'/dataelements/{result["data_elements"]["label"][0]}')
    #
    # assert "label" in dti
    # assert "description" in dti
    # assert "isAbout" in dti
    # assert "inProjects" in dti
    #
    # # make sure the text formatter doesn't fail horribly
    # rest_parser.setOutputFormat(RestParser.CLI_FORMAT)
    # txt = rest_parser.run(openneuro_files, f'/dataelements/{result["data_elements"]["label"][0]}')
    #

    dti = rest_parser.run(
        openneuro_files, "/dataelements/Left-WM-hypointensities Volume_mm3 (mm^3)"
    )
    print(dti)


def test_dataelement_details_in_projects_field(brain_vol_files: list[str]) -> None:
    rest_parser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    # result = rest_parser.run(openneuro_files, '/dataelements')
    # dti = rest_parser.run(openneuro_files, f'/dataelements/{result["data_elements"]["label"][0]}')
    # assert len(dti['inProjects']) >= 1

    # find a data element that we are using for at least one subject
    data_element_label = "Right-non-WM-hypointensities normMax (MR)"
    dti = rest_parser.run(brain_vol_files, f"/dataelements/{data_element_label}")
    assert len(dti["inProjects"]) >= 1



from dataclasses import dataclass
import json
from os.path import join
from pathlib import Path
import pandas as pd
import pytest
from nidm.experiment.Utils import map_variables_to_terms, write_json_mapping_file


@dataclass
class Setup:
    data: pd.DataFrame
    reproschema_json_map: dict
    bids_sidecar: dict
    bids_sidecar_simple: dict


@pytest.fixture(scope="module")
def setup() -> Setup:
    temp = {
        "participant_id": [
            "100",
            "101",
            "102",
            "103",
            "104",
            "105",
            "106",
            "107",
            "108",
            "109",
        ],
        "age": [18, 25, 30, 19, 35, 20, 27, 29, 38, 27],
        "sex": ["m", "m", "f", "m", "f", "f", "f", "f", "m", "m"],
    }

    data = pd.DataFrame(temp)

    reproschema_json_map = json.loads(
        """
        {
            "DD(source='participants.tsv', variable='participant_id')": {
                "label": "participant_id",
                "description": "subject/participant identifier",
                "source_variable": "participant_id",
                "responseOptions": {
                    "valueType": "http://www.w3.org/2001/XMLSchema#string"
                },
                "isAbout": [
                    {
                        "@id": "https://ndar.nih.gov/api/datadictionary/v2/dataelement/src_subject_id",
                        "label": "src_subject_id"
                    }
                ]
            },
            "DD(source='participants.tsv', variable='age')": {
                "responseOptions": {
                    "unitCode": "years",
                    "minValue": "0",
                    "maxValue": "100",
                    "valueType": "http://www.w3.org/2001/XMLSchema#integer"
                },
                "label": "age",
                "description": "age of participant",
                "source_variable": "age",
                "associatedWith": "NIDM",
                "isAbout": [
                    {
                        "@id": "http://uri.interlex.org/ilx_0100400",
                        "label": "Age"
                    }
                ]
            },
            "DD(source='participants.tsv', variable='sex')": {
                "responseOptions": {
                    "minValue": "NA",
                    "maxValue": "NA",
                    "unitCode": "NA",
                    "valueType": "http://www.w3.org/2001/XMLSchema#complexType",
                    "choices": {
                        "Male": "m",
                        "Female": "f"
                    }
                },
                "label": "sex",
                "description": "biological sex of participant",
                "source_variable": "sex",
                "associatedWith": "NIDM",
                "isAbout": [
                    {
                        "@id": "http://uri.interlex.org/ilx_0738439",
                        "label": "SEX"
                    }
                ]
            }
        }"""
    )

    bids_sidecar = json.loads(
        """
        {
            "age": {
                "label": "age",
                "description": "age of participant",
                "source_variable": "age",
                "associatedWith": "NIDM",
                "isAbout": [
                    {
                        "@id": "http://uri.interlex.org/ilx_0100400",
                        "label": "Age"
                    }
                ],
                "valueType": "http://www.w3.org/2001/XMLSchema#integer",
                "minValue": "10",
                "maxValue": "100"
            },
            "sex": {
                "minValue": "NA",
                "maxValue": "NA",
                "unitCode": "NA",
                "valueType": "http://www.w3.org/2001/XMLSchema#complexType",
                "levels": {
                    "Male": "m",
                    "Female": "f"
                },
                "label": "sex",
                "description": "biological sex of participant",
                "source_variable": "sex",
                "associatedWith": "NIDM",
                "isAbout": [
                    {
                        "@id": "http://uri.interlex.org/ilx_0738439",
                        "label": "SEX"
                    }
                ]
            }
        }
        """
    )
    bids_sidecar_simple = json.loads(
        """
        {
            "age": {
                "description": "age of participant"
                },
            "sex": {
                "description": "biological sex of participant"
            }
        }
        """
    )

    return Setup(
        data=data,
        reproschema_json_map=reproschema_json_map,
        bids_sidecar=bids_sidecar,
        bids_sidecar_simple=bids_sidecar_simple,
    )


def test_map_vars_to_terms_BIDS(setup: Setup, tmp_path: Path) -> None:
    """
    This function will test the Utils.py "map_vars_to_terms" function with a BIDS-formatted
    JSON sidecar file
    """

    # test BIDS sidecar json file with all pynidm annotations
    column_to_terms, cde = map_variables_to_terms(
        df=setup.data,
        json_source=setup.bids_sidecar,
        directory=str(tmp_path),
        assessment_name="test",
        bids=True,
    )

    # check whether JSON mapping structure returned from map_variables_to_terms matches the
    # reproshema structure
    assert "DD(source='test', variable='age')" in column_to_terms
    assert "DD(source='test', variable='sex')" in column_to_terms
    assert "isAbout" in column_to_terms["DD(source='test', variable='age')"]
    assert (
        "http://uri.interlex.org/ilx_0100400"
        == column_to_terms["DD(source='test', variable='age')"]["isAbout"][0]["@id"]
    )
    assert (
        "http://uri.interlex.org/ilx_0738439"
        == column_to_terms["DD(source='test', variable='sex')"]["isAbout"][0]["@id"]
    )
    assert (
        "responseOptions" in column_to_terms["DD(source='test', variable='sex')"].keys()
    )
    assert (
        "choices"
        in column_to_terms["DD(source='test', variable='sex')"][
            "responseOptions"
        ].keys()
    )
    assert (
        "Male"
        in column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ].keys()
    )
    assert (
        "m"
        == column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ]["Male"]
    )
    assert (
        "Male"
        in column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ].keys()
    )
    assert (
        "m"
        == column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ]["Male"]
    )

    # force writing of column_to_terms structure because here we're not doing annotations and so
    # map_variables_to_terms won't write it out since we supplied one for it to open...thus it already exists
    # and no annotations were made so it should exist in its original form.
    # By explicitly writing it out here, after running map_variables_to_terms, we can assure it's the same as the
    # original.

    # write annotations to json file since data element annotations are complete
    write_json_mapping_file(
        column_to_terms, join(str(tmp_path), "nidm_annotations.json"), True
    )

    # now check the JSON sidecar file created by map_variables_to_terms which should match BIDS format
    with open(tmp_path / "nidm_annotations.json", encoding="utf-8") as fp:
        bids_sidecar = json.load(fp)

    assert "age" in bids_sidecar.keys()
    assert "sex" in bids_sidecar.keys()
    assert "isAbout" in bids_sidecar["age"].keys()
    assert (
        "http://uri.interlex.org/ilx_0100400"
        == bids_sidecar["age"]["isAbout"][0]["@id"]
    )
    assert (
        "http://uri.interlex.org/ilx_0738439"
        == bids_sidecar["sex"]["isAbout"][0]["@id"]
    )
    assert "levels" in bids_sidecar["sex"].keys()
    assert "Male" in bids_sidecar["sex"]["levels"].keys()
    assert "m" == bids_sidecar["sex"]["levels"]["Male"]
    assert "Male" in bids_sidecar["sex"]["levels"].keys()
    assert "m" == bids_sidecar["sex"]["levels"]["Male"]

    # check the CDE dataelement graph for correct information
    query = """
        prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>

        select distinct ?uuid ?DataElements ?property ?value
            where {

                ?uuid a/rdfs:subClassOf* nidm:DataElement ;
                    ?property ?value .

        }"""
    qres = cde.query(query)

    results = []
    for row in qres:
        results.append(list(row))

    assert len(results) == 20


def test_map_vars_to_terms_BIDS_simple(setup: Setup, tmp_path: Path) -> None:
    """
    This function will test the Utils.py "map_vars_to_terms" function with a BIDS-formatted
    JSON sidecar file
    """

    # test BIDS sidecar json file with all pynidm annotations
    column_to_terms, cde = map_variables_to_terms(
        df=setup.data,
        json_source=setup.bids_sidecar_simple,
        directory=str(tmp_path),
        assessment_name="test",
        associate_concepts=False,
        bids=True,
    )

    # check whether JSON mapping structure returned from map_variables_to_terms matches the
    # reproshema structure
    assert "DD(source='test', variable='age')" in column_to_terms
    assert "DD(source='test', variable='sex')" in column_to_terms
    assert "description" in column_to_terms["DD(source='test', variable='age')"]
    assert "description" in column_to_terms["DD(source='test', variable='sex')"]

    # force writing of column_to_terms structure because here we're not doing annotations and so
    # map_variables_to_terms won't write it out since we supplied one for it to open...thus it already exists
    # and no annotations were made so it should exist in its original form.
    # By explicitly writing it out here, after running map_variables_to_terms, we can assure it's the same as the
    # original.

    # write annotations to json file since data element annotations are complete
    write_json_mapping_file(
        column_to_terms, join(str(tmp_path), "nidm_annotations.json"), True
    )

    # now check the JSON sidecar file created by map_variables_to_terms which should match BIDS format
    with open(tmp_path / "nidm_annotations.json", encoding="utf-8") as fp:
        bids_sidecar = json.load(fp)

    assert "age" in bids_sidecar.keys()
    assert "sex" in bids_sidecar.keys()
    assert "description" in bids_sidecar["age"]
    assert "description" in bids_sidecar["sex"]

    # check the CDE dataelement graph for correct information
    query = """
        prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>

        select distinct ?uuid ?DataElements ?property ?value
            where {

                ?uuid a/rdfs:subClassOf* nidm:DataElement ;
                    ?property ?value .

        }"""
    qres = cde.query(query)

    results = []
    for row in qres:
        results.append(list(row))

    assert len(results) == 16


def test_map_vars_to_terms_reproschema(setup: Setup, tmp_path: Path) -> None:
    """
    This function will test the Utils.py "map_vars_to_terms" function with a reproschema-formatted
    JSON sidecar file
    """

    column_to_terms, cde = map_variables_to_terms(
        df=setup.data,
        json_source=setup.reproschema_json_map,
        directory=str(tmp_path),
        assessment_name="test",
    )

    # check whether JSON mapping structure returned from map_variables_to_terms matches the
    # reproshema structure
    assert "DD(source='test', variable='age')" in column_to_terms
    assert "DD(source='test', variable='sex')" in column_to_terms
    assert "isAbout" in column_to_terms["DD(source='test', variable='age')"]
    assert (
        "http://uri.interlex.org/ilx_0100400"
        == column_to_terms["DD(source='test', variable='age')"]["isAbout"][0]["@id"]
    )
    assert (
        "http://uri.interlex.org/ilx_0738439"
        == column_to_terms["DD(source='test', variable='sex')"]["isAbout"][0]["@id"]
    )
    assert (
        "responseOptions" in column_to_terms["DD(source='test', variable='sex')"].keys()
    )
    assert (
        "choices"
        in column_to_terms["DD(source='test', variable='sex')"][
            "responseOptions"
        ].keys()
    )
    assert (
        "Male"
        in column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ].keys()
    )
    assert (
        "m"
        == column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ]["Male"]
    )
    assert (
        "Male"
        in column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ].keys()
    )
    assert (
        "m"
        == column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ]["Male"]
    )

    # force writing of column_to_terms structure because here we're not doing annotations and so
    # map_variables_to_terms won't write it out since we supplied one for it to open...thus it already exists
    # and no annotations were made so it should exist in its original form.
    # By explicitly writing it out here, after running map_variables_to_terms, we can assure it's the same as the
    # original.

    # write annotations to json file since data element annotations are complete
    write_json_mapping_file(
        column_to_terms, join(str(tmp_path), "nidm_annotations.json"), False
    )

    # now check the JSON mapping file created by map_variables_to_terms which should match Reproschema format
    with open(tmp_path / "nidm_annotations_annotations.json", encoding="utf-8") as fp:
        json.load(fp)

    assert "DD(source='test', variable='age')" in column_to_terms
    assert "DD(source='test', variable='sex')" in column_to_terms
    assert "isAbout" in column_to_terms["DD(source='test', variable='age')"]
    assert (
        "http://uri.interlex.org/ilx_0100400"
        == column_to_terms["DD(source='test', variable='age')"]["isAbout"][0]["@id"]
    )
    assert (
        "http://uri.interlex.org/ilx_0738439"
        == column_to_terms["DD(source='test', variable='sex')"]["isAbout"][0]["@id"]
    )
    assert (
        "responseOptions" in column_to_terms["DD(source='test', variable='sex')"].keys()
    )
    assert (
        "choices"
        in column_to_terms["DD(source='test', variable='sex')"][
            "responseOptions"
        ].keys()
    )
    assert (
        "Male"
        in column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ].keys()
    )
    assert (
        "m"
        == column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ]["Male"]
    )
    assert (
        "Male"
        in column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ].keys()
    )
    assert (
        "m"
        == column_to_terms["DD(source='test', variable='sex')"]["responseOptions"][
            "choices"
        ]["Male"]
    )

    # check the CDE dataelement graph for correct information
    query = """
        prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>

        select distinct ?uuid ?DataElements ?property ?value
            where {

                ?uuid a/rdfs:subClassOf* nidm:DataElement ;
                    ?property ?value .

        }"""
    qres = cde.query(query)

    results = []
    for row in qres:
        results.append(list(row))

    assert len(results) == 20



import csv
import os
import sys
import tempfile
import warnings
import click
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.metrics import (
    calinski_harabaz_score,
    davies_bouldin_score,
    silhouette_score,
)
from nidm.experiment.Query import GetProjectsUUID
from nidm.experiment.tools.click_base import cli
from nidm.experiment.tools.rest import RestParser
from .utils import Reporter


@cli.command()
@click.option(
    "--nidm_file_list",
    "-nl",
    required=True,
    help="A comma separated list of NIDM files with full path",
)
@click.option(
    "--var",
    "-variables",
    required=True,
    help='This parameter is for the variables the user would like to complete the k-means algorithm on.\nThe way this looks in the command is python3 nidm_kmeans.py -nl MTdemog_aseg_v2.ttl -v "fs_003343,age*sex,sex,age,group,age*group,bmi"',
)
@click.option(
    "--k_range",
    "-k",
    required=True,
    help="The maximum number of clusters to try. The algorithm will go from 2 to this number to determine the optimal number of clusters.",
)
@click.option(
    "--optimal_cluster_method",
    "-m",
    required=True,
    help="The criterion used to select the optimal partitioning (either Gap Statistic, Elbow Method, Silhouette Coefficient, Calinski-Harabasz Index, or Davies_Bouldin Index).",
)
@click.option(
    "--output_file",
    "-o",
    required=False,
    help="Optional output file (TXT) to store results of the linear regression, contrast, and regularization",
)
def k_means(nidm_file_list, output_file, var, k_range, optimal_cluster_method):
    """
    This function provides a tool to complete k-means clustering on NIDM data.
    """
    global v  # Needed to do this because the code only used the parameters in the first method, meaning I had to move it all to method 1.
    v = (
        var.strip()
    )  # used in data_aggregation, kmenas(), spaces stripped from left and right
    with Reporter(output_file) as reporter:
        global n  # used in data_aggregation()
        n = nidm_file_list
        global k_num
        k_num = int(k_range)
        global cm
        cm = optimal_cluster_method
        data_aggregation(reporter)
        dataparsing(reporter)
        cluster_number()


def data_aggregation(reporter):  # all data from all the files is collected
    """This function provides query support for NIDM graphs."""
    # if there is a CDE file list, seed the CDE cache
    if v:  # ex: age,sex,DX_GROUP
        print("*" * 107)
        command = (
            "pynidm k-means -nl "
            + n
            + ' -variables "'
            + v
            + '" '
            + "-k "
            + str(k_num)
            + " -m "
            + cm
        )

        reporter.print("Your command was:", command)
        verbosity = 0
        restParser = RestParser(verbosity_level=int(verbosity))
        restParser.setOutputFormat(RestParser.OBJECT_FORMAT)
        global df_list  # used in dataparsing()
        df_list = []
        # set up uri to do fields query for each nidm file
        global file_list
        file_list = n.split(",")
        df_list_holder = {}
        for i in range(len(file_list)):
            df_list_holder[i] = []
        df_holder = {}
        for i in range(len(file_list)):
            df_holder[i] = []
        global condensed_data_holder
        condensed_data_holder = {}
        for i in range(len(file_list)):
            condensed_data_holder[i] = []

        count = 0
        not_found_count = 0
        for nidm_file in file_list:
            # get project UUID
            project = GetProjectsUUID([nidm_file])
            # split the model into its constituent variables
            global var_list
            # below, we edit the model so it splits by +,~, or =. However, to help it out in catching everything
            # we replaced ~ and = with a + so that we can still use split. Regex wasn't working.
            var_list = [vv.strip() for vv in v.split(",")]
            # set the dependent variable to the one dependent variable in the model
            global variables  # used in dataparsing()
            variables = ""
            for vr in reversed(var_list):
                if (
                    "*" not in vr
                ):  # removing the star term from the columns we're about to pull from data
                    variables += vr + ","
                else:
                    print(
                        "Interacting variables are not present in clustering models. They will be removed."
                    )
            variables = variables[:-1]
            uri = (
                "/projects/"
                + project[0].toPython().split("/")[-1]
                + "?fields="
                + variables
            )
            # get fields output from each file and concatenate
            df_list_holder[count].append(pd.DataFrame(restParser.run([nidm_file], uri)))
            # global dep_var
            df = pd.concat(df_list_holder[count])
            with tempfile.NamedTemporaryFile(
                delete=False
            ) as temp:  # turns the dataframe into a temporary csv
                df.to_csv(temp.name + ".csv")
                temp.close()
            with open(temp.name + ".csv", encoding="utf-8") as fp:
                data = list(
                    csv.reader(fp)
                )  # makes the csv a 2D list to make it easier to call the contents of certain cells

            var_list = variables.split(",")  # makes a list of the independent variables
            numcols = (len(data) - 1) // (
                len(var_list)
            )  # Finds the number of columns in the original dataframe
            condensed_data_holder[count] = [
                [0] * (len(var_list))
            ]  # makes an array 1 row by the number of necessary columns
            for _ in range(
                numcols
            ):  # makes the 2D array big enough to store all of the necessary values in the edited dataset
                condensed_data_holder[count].append([0] * (len(var_list)))
            var_list = [vr.split("/")[-1] for vr in var_list]
            for i, vr in enumerate(var_list):
                # stores the independent variable names in the first row
                condensed_data_holder[count][0][i] = vr
            numrows = 1  # begins at the first row to add data
            fieldcolumn = (
                0  # the column the variable name is in in the original dataset
            )
            valuecolumn = 0  # the column the value is in in the original dataset
            datacolumn = 0  # if it is identified by the dataElement name instead of the field's name
            not_found_list = []
            for i in range(len(data[0])):
                if (
                    data[0][i] == "sourceVariable"
                ):  # finds the column where the variable names are
                    fieldcolumn = i
                elif (
                    data[0][i] == "source_variable"
                ):  # finds the column where the variable names are
                    fieldcolumn = i
                elif data[0][i] == "isAbout":
                    aboutcolumn = i
                elif data[0][i] == "label":
                    namecolumn = i  # finds the column where the variable names are
                elif data[0][i] == "value":
                    valuecolumn = i  # finds the column where the values are
                elif (
                    data[0][i] == "dataElement"
                ):  # finds the column where the data element is if necessary
                    datacolumn = i
            for i in range(
                len(condensed_data_holder[count][0])
            ):  # starts iterating through the dataset, looking for the name in that
                for j in range(
                    1, len(data)
                ):  # column, so it can append the values under the proper variables
                    try:
                        if (
                            data[j][fieldcolumn] == condensed_data_holder[count][0][i]
                        ):  # in the dataframe, the name is in column 3
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                        elif data[j][aboutcolumn] == condensed_data_holder[count][0][i]:
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                        elif (
                            condensed_data_holder[count][0][i] in data[j][aboutcolumn]
                        ):  # this is in case the uri only works by querying the part after the last backslash
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                        elif (
                            data[j][namecolumn] == condensed_data_holder[count][0][i]
                        ):  # in the dataframe, the name is in column 12
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                        elif (
                            condensed_data_holder[count][0][i] == data[j][datacolumn]
                        ):  # in the dataframe, the name is in column 9
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                    except IndexError:
                        numrows = numrows + 1
                numrows = 1  # resets to the first row for the next variable
            temp_list = condensed_data_holder[count]
            for j in range(
                len(temp_list[0]) - 1, 0, -1
            ):  # if the software appends a column with 0 as the heading, it removes this null column
                if temp_list[0][j] == "0" or temp_list[0][j] == "NaN":
                    for row in condensed_data_holder[count]:
                        row.pop(j)
            rowsize = len(condensed_data_holder[count][0])
            count1 = 0
            for i in range(0, rowsize):
                for row in condensed_data_holder[count]:
                    if row[i] == 0 or row[i] == "NaN" or row[i] == "0":
                        count1 = count1 + 1
                if count1 > len(condensed_data_holder[count]) - 2:
                    not_found_list.append(condensed_data_holder[count][0][i])
                count1 = 0
            for i, cdh in enumerate(condensed_data_holder[count][0]):
                if " " in cdh:
                    condensed_data_holder[count][0][i] = cdh.replace(" ", "_")
            var_list = [vr.split("/")[-1].replace(" ", "_") for vr in var_list]
            count += 1
            if len(not_found_list) > 0:
                print("*" * 107)
                print()
                reporter.print("Your variables were", v)
                reporter.print()
                reporter.print(
                    "The following variables were not found in "
                    + nidm_file
                    + ". The model cannot run because this will skew the data. Try checking your spelling or use nidm_query.py to see other possible variables."
                )
                for i, nf in enumerate(not_found_list):
                    reporter.print(f"{i + 1}. {nf}")
                not_found_list.clear()
                not_found_count += 1
                print()
        if not_found_count > 0:
            sys.exit(1)

    else:
        print("ERROR: No query parameter provided.  See help:")
        print()
        os.system("pynidm k-means --help")
        sys.exit(1)


def dataparsing(
    reporter,
):  # The data is changed to a format that is usable by the linear regression method
    global condensed_data
    condensed_data = []
    for i in range(0, len(file_list)):
        condensed_data += condensed_data_holder[i]
    global k_num
    if len(condensed_data[0]) <= k_num:
        print(
            "\nThe maximum number of clusters specified is greater than the amount of data present."
        )
        print(
            "The algorithm cannot run with this, so k_num will be reduced to 1 less than the length of the dataset."
        )
        k_num = len(condensed_data) - 1
        print("The k_num value is now:", k_num)
    x = pd.read_csv(
        opencsv(condensed_data)
    )  # changes the dataframe to a csv to make it easier to work with
    x.head()  # prints what the csv looks like
    x.dtypes  # checks data format
    obj_df = x.select_dtypes  # puts all the variables in a dataset
    x.shape  # says number of rows and columns in form of tuple
    x.describe()  # says dataset statistics
    obj_df = x.select_dtypes(
        include=["object"]
    ).copy()  # takes everything that is an object (not float or int) and puts it in a new dataset
    obj_df.head()  # prints the new dataset
    int_df = x.select_dtypes(
        include=["int64"]
    ).copy()  # takes everything that is an int and puts it in a new dataset
    float_df = x.select_dtypes(
        include=["float64"]
    ).copy()  # takes everything that is a float and puts it in a new dataset
    df_int_float = pd.concat([float_df, int_df], axis=1)
    stringvars = []  # starts a list that will store all variables that are not numbers
    for i in range(1, len(condensed_data)):  # goes through each variable
        for j in range(len(condensed_data[0])):  # in the 2D array
            try:  # if the value of the field can be turned into a float (is numerical)
                float(condensed_data[i][j])  # this means it's a number
            except ValueError:  # if it can't be (is a string)
                if (
                    condensed_data[0][j] not in stringvars
                ):  # adds the variable name to the list if it isn't there already
                    stringvars.append(condensed_data[0][j])
    le = (
        preprocessing.LabelEncoder()
    )  # anything involving le shows the encoding of categorical variables
    for sv in stringvars:
        le.fit(obj_df[sv].astype(str))
    obj_df_trf = obj_df.astype(str).apply(
        le.fit_transform
    )  # transforms the categorical variables into numbers.
    global df_final  # also used in linreg()
    if not obj_df_trf.empty:
        df_final = pd.concat(
            [df_int_float, obj_df_trf], axis=1
        )  # join_axes=[df_int_float.index])
    else:
        df_final = df_int_float
    reporter.print(df_final.to_string(header=True, index=True))
    reporter.print("\n" + ("*" * 107))
    reporter.print("\nModel Results: ")


def cluster_number():
    index = 0
    global levels  # also used in contrasting()
    levels = []
    for i in range(1, len(condensed_data)):
        if condensed_data[i][index] not in levels:
            levels.append(condensed_data[i][index])
    levels = list(range(len(levels)))

    # Beginning of the linear regression
    global X
    # global y
    # Unsure on how to proceed here with interacting variables, since I'm sure dmatrices won't work

    # scaler = MinMaxScaler()
    #
    # for i in range(len(model_list)):
    #     scaler.fit(df_final[[model_list[i]]])
    #     df_final[[model_list[i]]] = scaler.transform(df_final[[model_list[i]]])
    X = df_final[var_list]
    if "ga" in cm.lower():
        print("\n\nGap Statistic")
        gaps = np.zeros((len(range(2, int(k_num)))))
        global resulting_df
        resulting_df = pd.DataFrame({"clusterCount": [], "gap": []})
        global gap_index, k
        for gap_index, k in enumerate(range(2, int(k_num))):
            dispersion_results = np.zeros(3)  # make three random datasets
            for i in range(3):
                random_reference = np.random.random_sample(size=df_final.shape)
                km = KMeans(k)
                km.fit(random_reference)
                dispersion_results[i] = km.inertia_
            km = KMeans(k)
            km.fit(df_final)

            original_sse = km.inertia_
            global gap
            gap = np.log(np.mean(dispersion_results)) - np.log(original_sse)

            gaps[gap_index] = gap
        max_gap = gaps[0]
        optimal_i = 0
        for i in range(1, len(gaps)):
            if gaps[i] > max_gap:
                optimal_i = i
                max_gap = gaps[i]
        optimal_cluster = optimal_i + 2
        print(
            "Optimal number of clusters: " + str(optimal_cluster)
        )  # the optimal number of clusters for gap statistic
        km = KMeans(
            n_clusters=optimal_cluster,
            init="k-means++",
            max_iter=300,
            n_init=10,
            random_state=0,
        )
        labels = km.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap="viridis", zorder=2)
        ax.axis("equal")
        plt.show()

    if "el" in cm.lower():
        print("\n\nElbow Method")
        sse = []
        for i in range(2, int(k_num)):
            km = KMeans(
                n_clusters=i, init="k-means++", max_iter=300, n_init=10, random_state=0
            )
            model = km.fit(X)
            sse.append(km.inertia_)
        min_sse = sse[0]
        max_sse = sse[0]
        max_i = 0
        min_i = 0
        for i in range(1, len(sse)):
            if sse[i] >= max_sse:
                max_sse = sse[i]
                max_i = i
            elif sse[i] <= min_sse:
                min_sse = sse[i]
                min_i = i
        p1 = np.array([min_i, sse[min_i]])
        p2 = np.array([max_i, sse[max_i]])
        # the way I am doing the elbow method is as follows:
        # the different sse values form a curve like an L (like an exponential decay)
        # The elbow is the point furthest from a line connecting max and min
        # So I am calculating the distance, and the maximum distance from point to curve shows the optimal point
        # AKA the number of clusters
        dist = []
        for n, s in enumerate(sse):
            norm = np.linalg.norm
            p3 = np.array([n, s])
            dist.append(np.abs(norm(np.cross(p2 - p1, p1 - p3))) / norm(p2 - p1))
        max_dist = dist[0]
        optimal_cluster = 2
        for x in range(1, len(dist)):
            if dist[x] >= max_dist:
                max_dist = dist[x]
                optimal_cluster = x + 2
        print(
            "Optimal number of clusters: " + str(optimal_cluster)
        )  # the optimal number of clusters for elbow method
        km = KMeans(
            n_clusters=optimal_cluster,
            init="k-means++",
            max_iter=300,
            n_init=10,
            random_state=0,
        )
        labels = km.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap="viridis", zorder=2)
        ax.axis("equal")
        plt.show()

    if "si" in cm.lower():
        print("Silhouette Score\n")
        ss = []
        for i in range(2, int(k_num)):
            km = KMeans(
                n_clusters=i, init="k-means++", max_iter=300, n_init=10, random_state=0
            )
            cluster_labels = km.fit_predict(X)
            silhouette_avg = silhouette_score(X, cluster_labels)
            ss.append(silhouette_avg)
        optimal_i = 0
        distance_to_one = abs(1 - ss[0])
        for i, s in enumerate(ss):
            if abs(1 - s) <= distance_to_one:
                optimal_i = i
                distance_to_one = abs(1 - s)
        n_clusters = optimal_i + 2
        print(
            "Optimal number of clusters: " + str(n_clusters)
        )  # the optimal number of clusters
        km = KMeans(
            n_clusters=n_clusters,
            init="k-means++",
            max_iter=300,
            n_init=10,
            random_state=0,
        )
        labels = km.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap="viridis", zorder=2)
        ax.axis("equal")
        plt.show()

    if "ca" in cm.lower():
        warnings.filterwarnings(
            "ignore", category=FutureWarning
        )  # it is a function for 0.24 but says it is deprecated in 0.23
        print("Calinski-Harabasz Index\n")
        pca = PCA(n_components=2)
        impca = pca.fit_transform(X)
        scores = []

        centers = list(range(2, int(k_num)))
        for center in centers:
            km = KMeans(
                n_clusters=center,
                init="k-means++",
                max_iter=300,
                n_init=10,
                random_state=0,
            ).fit(impca)
            score = calinski_harabaz_score(impca, km.labels_)
            scores.append(score)
        optimal_i = 0
        max_score = scores[0]

        for i in range(1, len(scores)):
            if scores[i] >= max_score:
                optimal_i = i
                max_score = scores[i]
        n_clusters = optimal_i + 2
        print(
            "Optimal number of clusters: " + str(n_clusters)
        )  # the optimal number of clusters
        km = KMeans(
            n_clusters=n_clusters,
            init="k-means++",
            max_iter=300,
            n_init=10,
            random_state=0,
        )
        labels = km.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap="viridis", zorder=2)
        ax.axis("equal")
        plt.show()

    if "da" in cm.lower():
        print("Davies-Bouldin Index\n")
        scores = []

        centers = list(range(2, int(k_num)))
        for center in centers:
            km = KMeans(
                n_clusters=center,
                init="k-means++",
                max_iter=300,
                n_init=10,
                random_state=0,
            )
            model = km.fit_predict(X)
            score = davies_bouldin_score(X, model)
            scores.append(score)
        optimal_i = 0
        min_score = scores[0]

        for i in range(1, len(scores)):
            if scores[i] <= min_score:
                optimal_i = i
                min_score = scores[i]
        n_clusters = optimal_i + 2
        print(
            "Optimal number of clusters: " + str(n_clusters)
        )  # the optimal number of clusters
        km = KMeans(
            n_clusters=n_clusters,
            init="k-means++",
            max_iter=300,
            n_init=10,
            random_state=0,
        )
        labels = km.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap="viridis", zorder=2)
        ax.axis("equal")
        plt.show()

    if "de" in cm.lower():
        print("Dendrogram")
        # ask for help: how does one do a dendrogram, also without graphing?


def opencsv(data):
    """saves a list of lists as a csv and opens"""
    handle, fn = tempfile.mkstemp(suffix=".csv")
    with os.fdopen(
        handle, "w", encoding="utf8", errors="surrogateescape", newline=""
    ) as f:
        writer = csv.writer(f)
        writer.writerows(data)
    return fn


# it can be used calling the script `python nidm_query.py -nl ... -q ..
if __name__ == "__main__":
    k_means()



from __future__ import annotations
from pathlib import Path
from click.testing import CliRunner
import pytest
from nidm.experiment.tools.nidm_linreg import linear_regression


def test_simple_model(brain_vol_files: list[str]) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        r = runner.invoke(
            linear_regression,
            [
                "--nidm_file_list",
                ",".join(brain_vol_files),
                "--ml",
                "fs_000008 = DX_GROUP + http://uri.interlex.org/ilx_0100400",
                "-o",
                "output.txt",
            ],
        )
        assert r.exit_code == 0
        out = Path("output.txt").read_text(encoding="utf-8")

    # check if model was read correctly
    assert "fs_000008 ~ ilx_0100400 + DX_GROUP" in out

    # check correct number of observations
    assert "No. Observations:                  53" in out

    # check model coefficients
    assert (
        "const          27.7816      4.378      6.345      0.000      18.988      36.576"
        in out
    )
    assert (
        "ilx_0100400    -0.1832      0.173     -1.061      0.294      -0.530       0.164"
        in out
    )
    assert (
        "DX_GROUP        3.4908      4.031      0.866      0.391      -4.605      11.587"
        in out
    )


def test_model_with_contrasts(brain_vol_files: list[str]) -> None:
    # run linear regression tool with simple model and evaluate output

    runner = CliRunner()
    with runner.isolated_filesystem():
        r = runner.invoke(
            linear_regression,
            [
                "--nidm_file_list",
                ",".join(brain_vol_files),
                "--ml",
                "fs_000008 = DX_GROUP + http://uri.interlex.org/ilx_0100400",
                "--ctr",
                "DX_GROUP",
                "-o",
                "output.txt",
            ],
        )
        assert r.exit_code == 0
        out = Path("output.txt").read_text(encoding="utf-8")

    # check if model was read correctly
    assert "fs_000008 ~ ilx_0100400 + DX_GROUP" in out

    # check correct number of observations
    assert "No. Observations:                  53" in out

    # check model coefficients for different codings
    assert (
        "C(DX_GROUP, Treatment)[T.1]     0.7307      4.209      0.174      0.863      -7.727       9.189"
        in out
    )
    assert (
        "C(DX_GROUP, Treatment)[T.2]    32.7462     15.984      2.049      0.046       0.625      64.868"
        in out
    )
    assert (
        "C(DX_GROUP, Simple)[Simp.0]     0.7307      4.209      0.174      0.863      -7.727       9.189"
        in out
    )
    assert (
        "C(DX_GROUP, Simple)[Simp.1]    32.7462     15.984      2.049      0.046       0.625      64.868"
        in out
    )
    assert (
        "C(DX_GROUP, Sum)[S.0]   -11.1590      5.713     -1.953      0.057     -22.639       0.321"
        in out
    )
    assert (
        "C(DX_GROUP, Sum)[S.1]   -10.4283      5.631     -1.852      0.070     -21.743       0.887"
        in out
    )
    assert (
        "C(DX_GROUP, Diff)[D.0]     0.7307      4.209      0.174      0.863      -7.727       9.189"
        in out
    )
    assert (
        "C(DX_GROUP, Diff)[D.1]    32.0155     15.896      2.014      0.050       0.070      63.961"
        in out
    )
    assert (
        "C(DX_GROUP, Helmert)[H.1]     0.3653      2.104      0.174      0.863      -3.864       4.594"
        in out
    )
    assert (
        "C(DX_GROUP, Helmert)[H.2]    10.7936      5.267      2.049      0.046       0.209      21.378"
        in out
    )


@pytest.mark.skip(
    reason="regularization weights seem to be different depending on the platform"
)
def test_model_with_contrasts_reg_L1(brain_vol_files: list[str]) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        r = runner.invoke(
            linear_regression,
            [
                "--nidm_file_list",
                ",".join(brain_vol_files),
                "--ml",
                "fs_000008 = DX_GROUP + http://uri.interlex.org/ilx_0100400",
                "--ctr",
                "DX_GROUP",
                "--regularization",
                "L1",
                "-o",
                "output.txt",
            ],
        )
        assert r.exit_code == 0
        out = Path("output.txt").read_text(encoding="utf-8")

    # check if model was read correctly
    assert "fs_000008 ~ ilx_0100400 + DX_GROUP" in out

    # check correct number of observations
    assert "No. Observations:                  53" in out

    # check lasso regression
    assert "Alpha with maximum likelihood (range: 1 to 700) = 43.000000" in out
    assert "Current Model Score = 0.000000" in out
    assert "ilx_0100400 	 -0.000000" in out
    assert "DX_GROUP 	 0.000000" in out
    assert "Intercept: 26.000000" in out


@pytest.mark.skip(
    reason="regularization weights seem to be different depending on the platform"
)
def test_model_with_contrasts_reg_L2(brain_vol_files: list[str]) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        r = runner.invoke(
            linear_regression,
            [
                "--nidm_file_list",
                ",".join(brain_vol_files),
                "--ml",
                "fs_000008 = DX_GROUP + http://uri.interlex.org/ilx_0100400",
                "--ctr",
                "DX_GROUP",
                "--regularization",
                "L2",
                "-o",
                "output.txt",
            ],
        )
        assert r.exit_code == 0
        out = Path("output.txt").read_text(encoding="utf-8")

    # check if model was read correctly
    assert "fs_000008 ~ ilx_0100400 + DX_GROUP" in out

    # check correct number of observations
    assert "No. Observations:                  53" in out

    # check lasso regression
    assert "Alpha with maximum likelihood (range: 1 to 700) = 699.000000" in out
    assert "Current Model Score = 0.017618" in out
    assert "ilx_0100400 	 -0.148397" in out
    assert "DX_GROUP 	 0.071356" in out
    assert "Intercept: 28.951297" in out



import prov.model as pm
from .Core import Core, getUUID
from ..core import Constants


class AcquisitionObject(pm.ProvEntity, Core):
    """Class for NIDM-Experimenent AcquisitionObject-Level Objects.

    Default constructor uses empty graph with namespaces added from NIDM/Scripts/Constants.py.
    Additional alternate constructors for user-supplied graphs and default namespaces (i.e. from Constants.py)
    and user-supplied graph and namespaces

    @author: David Keator <dbkeator@uci.edu>
    @copyright: University of California, Irvine 2017

    """

    # constructor
    def __init__(self, acquisition, attributes=None, uuid=None):
        """
        Default constructor, creates an acquisition object and links to acquisition activity object

        :param acquisition: a Acquisition activity object
        :param attributes: optional attributes to add to entity
        :param uuid: optional uuid...used mostly for reading in existing NIDM document
        :return: none

        """

        # carry graph object around
        self.graph = acquisition.graph

        if uuid is None:
            self._uuid = getUUID()
            # execute default parent class constructor
            super().__init__(
                acquisition.graph,
                pm.QualifiedName(
                    pm.Namespace("niiri", Constants.NIIRI), self.get_uuid()
                ),
                attributes,
            )
        else:
            # check if niiri namespace is already defined
            niiri_ns = self.find_namespace_with_uri(uri=str(Constants.NIIRI))
            self._uuid = uuid

            if niiri_ns is False:
                super().__init__(
                    acquisition.graph,
                    pm.QualifiedName(
                        pm.Namespace("niiri", Constants.NIIRI), self.get_uuid()
                    ),
                    attributes,
                )
            else:
                super().__init__(
                    acquisition.graph,
                    pm.QualifiedName(niiri_ns, self.get_uuid()),
                    attributes,
                )

        acquisition.graph._add_record(self)

        # create link to acquisition activity
        acquisition.add_acquisition_object(self)

    def __str__(self):
        return "NIDM-Experiment AcquisitionObject Class"
